"""Tests for pipette state changes in the protocol_engine state store.

DEPRECATED: Testing PipetteStore independently of PipetteView is no longer helpful.
Try to add new tests to test_pipette_state.py, where they can be tested together,
treating PipetteState as a private implementation detail.
"""

import pytest

from opentrons_shared_data.pipette import pipette_definition
from opentrons_shared_data.pipette.types import (
    LiquidClasses as VolumeModes,
)
from opentrons_shared_data.pipette.types import (
    PipetteNameType,
)

from ..pipette_fixtures import get_default_nozzle_map
from .command_fixtures import create_comment_command
from opentrons.protocol_engine.actions import (
    SetPipetteMovementSpeedAction,
    SucceedCommandAction,
)
from opentrons.protocol_engine.resources.pipette_data_provider import (
    LoadedStaticPipetteData,
)
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.fluid_stack import FluidStack
from opentrons.protocol_engine.state.pipettes import (
    BoundingNozzlesOffsets,
    CurrentDeckPoint,
    PipetteBoundingBoxOffsets,
    PipetteState,
    PipetteStore,
    StaticPipetteConfig,
)
from opentrons.protocol_engine.types import (
    AspiratedFluid,
    CurrentAddressableArea,
    CurrentWell,
    DeckPoint,
    FlowRates,
    FluidKind,
    LabwareWellId,
    LoadedPipette,
    TipGeometry,
)
from opentrons.types import MountType, Point


@pytest.fixture
def available_sensors() -> pipette_definition.AvailableSensorDefinition:
    """Provide a list of sensors."""
    return pipette_definition.AvailableSensorDefinition(
        sensors=["pressure", "capacitive", "environment"]
    )


@pytest.fixture
def subject() -> PipetteStore:
    """Get a PipetteStore test subject for all subsequent tests."""
    return PipetteStore()


def test_sets_initial_state(subject: PipetteStore) -> None:
    """It should initialize its state object properly."""
    result = subject.state

    assert result == PipetteState(
        pipettes_by_id={},
        pipette_contents_by_id={},
        current_location=None,
        current_deck_point=CurrentDeckPoint(mount=None, deck_point=None),
        attached_tip_by_id={},
        movement_speed_by_id={},
        static_config_by_id={},
        flow_rates_by_id={},
        nozzle_configuration_by_id={},
        liquid_presence_detection_by_id={},
        ready_to_aspirate_by_id={},
        has_clean_tips_by_id={},
        tip_source_by_id={},
    )


def test_location_state_update(subject: PipetteStore) -> None:
    """It should update pipette locations."""
    # Update the location to a well:
    dummy_command = create_comment_command()
    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_location=update_types.PipetteLocationUpdate(
                    pipette_id="pipette-id",
                    new_location=LabwareWellId(
                        labware_id="come on barbie",
                        well_name="let's go party",
                    ),
                    new_deck_point=DeckPoint(x=111, y=222, z=333),
                ),
                loaded_pipette=update_types.LoadPipetteUpdate(
                    pipette_id="pipette-id",
                    liquid_presence_detection=None,
                    pipette_name=PipetteNameType.P300_SINGLE,
                    mount=MountType.RIGHT,
                ),
            ),
        )
    )
    assert subject.state.current_location == CurrentWell(
        pipette_id="pipette-id", labware_id="come on barbie", well_name="let's go party"
    )
    assert subject.state.current_deck_point == CurrentDeckPoint(
        mount=MountType.RIGHT, deck_point=DeckPoint(x=111, y=222, z=333)
    )

    # Update the location to an addressable area:
    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_location=update_types.PipetteLocationUpdate(
                    pipette_id="pipette-id",
                    new_location=update_types.AddressableArea(
                        addressable_area_name="na na na na na"
                    ),
                    new_deck_point=DeckPoint(x=333, y=444, z=555),
                )
            ),
        )
    )
    assert subject.state.current_location == CurrentAddressableArea(
        pipette_id="pipette-id", addressable_area_name="na na na na na"
    )
    assert subject.state.current_deck_point == CurrentDeckPoint(
        mount=MountType.RIGHT, deck_point=DeckPoint(x=333, y=444, z=555)
    )

    # Clear the logical location:
    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_location=update_types.PipetteLocationUpdate(
                    pipette_id="pipette-id",
                    new_location=None,
                    new_deck_point=update_types.NO_CHANGE,
                )
            ),
        )
    )
    assert subject.state.current_location is None
    assert subject.state.current_deck_point == CurrentDeckPoint(
        mount=MountType.RIGHT, deck_point=DeckPoint(x=333, y=444, z=555)
    )

    # Repopulate the locations, then test clearing all pipette locations:
    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_location=update_types.PipetteLocationUpdate(
                    pipette_id="pipette-id",
                    new_location=update_types.AddressableArea(
                        addressable_area_name="na na na na na"
                    ),
                    new_deck_point=DeckPoint(x=333, y=444, z=555),
                )
            ),
        )
    )
    assert subject.state.current_location is not None
    assert subject.state.current_deck_point != CurrentDeckPoint(
        mount=None, deck_point=None
    )
    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(pipette_location=update_types.CLEAR),
        )
    )
    assert subject.state.current_location is None
    assert subject.state.current_deck_point == CurrentDeckPoint(
        mount=None, deck_point=None
    )


def test_handles_load_pipette(
    subject: PipetteStore,
    supported_tip_fixture: pipette_definition.SupportedTipsDefinition,
    available_sensors: pipette_definition.AvailableSensorDefinition,
) -> None:
    """It should add the pipette data to the state."""
    dummy_command = create_comment_command()

    load_pipette_update = update_types.LoadPipetteUpdate(
        pipette_id="pipette-id",
        pipette_name=PipetteNameType.P300_SINGLE,
        mount=MountType.LEFT,
        liquid_presence_detection=None,
    )

    config = LoadedStaticPipetteData(
        model="pipette-model",
        display_name="pipette name",
        min_volume=1.23,
        max_volume=4.56,
        channels=7,
        flow_rates=FlowRates(
            default_aspirate={"a": 1},
            default_dispense={"b": 2},
            default_blow_out={"c": 3},
        ),
        tip_configuration_lookup_table={4: supported_tip_fixture},
        nominal_tip_overlap={"default": 5},
        home_position=8.9,
        nozzle_offset_z=10.11,
        nozzle_map=get_default_nozzle_map(PipetteNameType.P300_SINGLE),
        back_left_corner_offset=Point(x=1, y=2, z=3),
        front_right_corner_offset=Point(x=4, y=5, z=6),
        pipette_lld_settings={},
        plunger_positions={
            "top": 0.0,
            "bottom": 5.0,
            "blow_out": 19.0,
            "drop_tip": 20.0,
        },
        shaft_ul_per_mm=5.0,
        available_sensors=available_sensors,
        volume_mode=VolumeModes.default,
        available_volume_modes_min_vol={},
    )
    config_update = update_types.PipetteConfigUpdate(
        pipette_id="pipette-id",
        config=config,
        serial_number="pipette-serial",
    )
    contents_update = update_types.PipetteUnknownFluidUpdate(pipette_id="pipette-id")

    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                loaded_pipette=load_pipette_update,
                pipette_config=config_update,
                pipette_aspirated_fluid=contents_update,
            ),
        )
    )

    result = subject.state

    assert result.pipettes_by_id["pipette-id"] == LoadedPipette(
        id="pipette-id",
        pipetteName=PipetteNameType.P300_SINGLE,
        mount=MountType.LEFT,
    )
    assert result.pipette_contents_by_id["pipette-id"] is None
    assert result.movement_speed_by_id["pipette-id"] is None
    assert result.attached_tip_by_id["pipette-id"] is None


def test_handles_pick_up_and_drop_tip(subject: PipetteStore) -> None:
    """It should set tip and volume details on pick up and drop tip."""
    dummy_command = create_comment_command()

    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                loaded_pipette=update_types.LoadPipetteUpdate(
                    pipette_id="abc",
                    pipette_name=PipetteNameType.P300_SINGLE,
                    mount=MountType.LEFT,
                    liquid_presence_detection=None,
                ),
                pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                    pipette_id="abc"
                ),
            ),
        )
    )

    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_tip_state=update_types.PipetteTipStateUpdate(
                    pipette_id="abc",
                    tip_geometry=TipGeometry(volume=42, length=101, diameter=8.0),
                    tip_source=LabwareWellId(labware_id="xyz", well_name="123"),
                ),
                pipette_aspirated_fluid=update_types.PipetteEmptyFluidUpdate(
                    pipette_id="abc", clean_tip=True
                ),
            ),
        )
    )
    assert subject.state.attached_tip_by_id["abc"] == TipGeometry(
        volume=42, length=101, diameter=8.0
    )
    assert subject.state.pipette_contents_by_id["abc"] == FluidStack()
    assert subject.state.tip_source_by_id["abc"] == LabwareWellId(
        labware_id="xyz", well_name="123"
    )

    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_tip_state=update_types.PipetteTipStateUpdate(
                    pipette_id="abc",
                    tip_geometry=None,
                    tip_source=None,
                ),
                pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                    pipette_id="abc"
                ),
            ),
        )
    )
    assert subject.state.attached_tip_by_id["abc"] is None
    assert subject.state.pipette_contents_by_id["abc"] is None


def test_handles_drop_tip_in_place(subject: PipetteStore) -> None:
    """It should clear tip and volume details after a drop tip in place."""
    dummy_command = create_comment_command()

    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                loaded_pipette=update_types.LoadPipetteUpdate(
                    pipette_id="xyz",
                    pipette_name=PipetteNameType.P300_SINGLE,
                    mount=MountType.LEFT,
                    liquid_presence_detection=None,
                ),
                pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                    pipette_id="xyz"
                ),
            ),
        )
    )
    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_tip_state=update_types.PipetteTipStateUpdate(
                    pipette_id="xyz",
                    tip_geometry=TipGeometry(volume=42, length=101, diameter=8.0),
                    tip_source=LabwareWellId(labware_id="abc", well_name="123"),
                ),
                pipette_aspirated_fluid=update_types.PipetteEmptyFluidUpdate(
                    pipette_id="xyz", clean_tip=False
                ),
            ),
        )
    )
    assert subject.state.attached_tip_by_id["xyz"] == TipGeometry(
        volume=42, length=101, diameter=8.0
    )
    assert subject.state.pipette_contents_by_id["xyz"] == FluidStack()
    assert subject.state.tip_source_by_id["xyz"] == LabwareWellId(
        labware_id="abc", well_name="123"
    )

    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_tip_state=update_types.PipetteTipStateUpdate(
                    pipette_id="xyz",
                    tip_geometry=None,
                    tip_source=None,
                ),
                pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                    pipette_id="xyz"
                ),
            ),
        )
    )
    assert subject.state.attached_tip_by_id["xyz"] is None
    assert subject.state.pipette_contents_by_id["xyz"] is None


def test_handles_unsafe_drop_tip_in_place(subject: PipetteStore) -> None:
    """It should clear tip and volume details after a drop tip in place."""
    dummy_command = create_comment_command()
    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                loaded_pipette=update_types.LoadPipetteUpdate(
                    pipette_id="xyz",
                    pipette_name=PipetteNameType.P300_SINGLE,
                    mount=MountType.LEFT,
                    liquid_presence_detection=None,
                ),
                pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                    pipette_id="xyz"
                ),
            ),
        )
    )
    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_tip_state=update_types.PipetteTipStateUpdate(
                    pipette_id="xyz",
                    tip_geometry=TipGeometry(volume=42, length=101, diameter=8.0),
                    tip_source=LabwareWellId(labware_id="abc", well_name="123"),
                ),
                pipette_aspirated_fluid=update_types.PipetteEmptyFluidUpdate(
                    pipette_id="xyz", clean_tip=False
                ),
            ),
        )
    )
    assert subject.state.attached_tip_by_id["xyz"] == TipGeometry(
        volume=42, length=101, diameter=8.0
    )
    assert subject.state.pipette_contents_by_id["xyz"] == FluidStack()
    assert subject.state.tip_source_by_id["xyz"] == LabwareWellId(
        labware_id="abc", well_name="123"
    )

    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_tip_state=update_types.PipetteTipStateUpdate(
                    pipette_id="xyz",
                    tip_geometry=None,
                    tip_source=None,
                ),
                pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                    pipette_id="xyz"
                ),
            ),
        )
    )
    assert subject.state.attached_tip_by_id["xyz"] is None
    assert subject.state.pipette_contents_by_id["xyz"] is None


def test_aspirate_adds_volume(subject: PipetteStore) -> None:
    """It should add volume to pipette after an aspirate."""
    dummy_command = create_comment_command()
    aspirate_update = update_types.StateUpdate(
        pipette_aspirated_fluid=update_types.PipetteAspiratedFluidUpdate(
            pipette_id="pipette-id",
            fluid=AspiratedFluid(kind=FluidKind.LIQUID, volume=42),
        )
    )

    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                loaded_pipette=update_types.LoadPipetteUpdate(
                    pipette_id="pipette-id",
                    pipette_name=PipetteNameType.P300_SINGLE,
                    mount=MountType.LEFT,
                    liquid_presence_detection=None,
                ),
                pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                    pipette_id="pipette-id"
                ),
            ),
        )
    )
    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_tip_state=update_types.PipetteTipStateUpdate(
                    pipette_id="pipette-id",
                    tip_geometry=TipGeometry(volume=42, length=101, diameter=8.0),
                    tip_source=LabwareWellId(labware_id="abc", well_name="123"),
                ),
                pipette_aspirated_fluid=update_types.PipetteEmptyFluidUpdate(
                    pipette_id="pipette-id", clean_tip=True
                ),
            ),
        )
    )

    subject.handle_action(
        SucceedCommandAction(command=dummy_command, state_update=aspirate_update)
    )

    assert subject.state.pipette_contents_by_id["pipette-id"] == FluidStack(
        _fluid_stack=[AspiratedFluid(kind=FluidKind.LIQUID, volume=42)]
    )

    subject.handle_action(
        SucceedCommandAction(command=dummy_command, state_update=aspirate_update)
    )

    assert subject.state.pipette_contents_by_id["pipette-id"] == FluidStack(
        _fluid_stack=[AspiratedFluid(kind=FluidKind.LIQUID, volume=84)]
    )


def test_dispense_subtracts_volume(subject: PipetteStore) -> None:
    """It should subtract volume from pipette after a dispense."""
    dummy_command = create_comment_command()
    dispense_update = update_types.StateUpdate(
        pipette_aspirated_fluid=update_types.PipetteEjectedFluidUpdate(
            pipette_id="pipette-id", volume=21
        )
    )

    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                loaded_pipette=update_types.LoadPipetteUpdate(
                    pipette_id="pipette-id",
                    pipette_name=PipetteNameType.P300_SINGLE,
                    mount=MountType.LEFT,
                    liquid_presence_detection=None,
                ),
                pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                    pipette_id="pipette-id"
                ),
            ),
        )
    )
    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_tip_state=update_types.PipetteTipStateUpdate(
                    pipette_id="pipette-id",
                    tip_geometry=TipGeometry(volume=47, length=101, diameter=8.0),
                    tip_source=LabwareWellId(labware_id="abc", well_name="123"),
                ),
                pipette_aspirated_fluid=update_types.PipetteEmptyFluidUpdate(
                    pipette_id="pipette-id", clean_tip=True
                ),
            ),
        )
    )
    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_aspirated_fluid=update_types.PipetteAspiratedFluidUpdate(
                    pipette_id="pipette-id",
                    fluid=AspiratedFluid(kind=FluidKind.LIQUID, volume=42),
                )
            ),
        )
    )
    subject.handle_action(
        SucceedCommandAction(command=dummy_command, state_update=dispense_update)
    )

    assert subject.state.pipette_contents_by_id["pipette-id"] == FluidStack(
        _fluid_stack=[AspiratedFluid(kind=FluidKind.LIQUID, volume=21)]
    )

    subject.handle_action(
        SucceedCommandAction(command=dummy_command, state_update=dispense_update)
    )

    assert subject.state.pipette_contents_by_id["pipette-id"] == FluidStack()


def test_blow_out_clears_volume(subject: PipetteStore) -> None:
    """It should wipe out the aspirated volume after a blowOut."""
    dummy_command = create_comment_command()

    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                loaded_pipette=update_types.LoadPipetteUpdate(
                    pipette_id="pipette-id",
                    pipette_name=PipetteNameType.P300_SINGLE,
                    mount=MountType.LEFT,
                    liquid_presence_detection=None,
                )
            ),
        )
    )
    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_tip_state=update_types.PipetteTipStateUpdate(
                    pipette_id="pipette-id",
                    tip_geometry=TipGeometry(volume=47, length=101, diameter=8.0),
                    tip_source=LabwareWellId(labware_id="abc", well_name="123"),
                ),
                pipette_aspirated_fluid=update_types.PipetteEmptyFluidUpdate(
                    pipette_id="pipette-id", clean_tip=True
                ),
            ),
        )
    )
    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_aspirated_fluid=update_types.PipetteAspiratedFluidUpdate(
                    pipette_id="pipette-id",
                    fluid=AspiratedFluid(kind=FluidKind.LIQUID, volume=42),
                )
            ),
        )
    )
    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                    pipette_id="pipette-id"
                )
            ),
        )
    )

    assert subject.state.pipette_contents_by_id["pipette-id"] is None


def test_set_movement_speed(subject: PipetteStore) -> None:
    """It should issue an action to set the movement speed."""
    subject.handle_action(
        SetPipetteMovementSpeedAction(pipette_id="pipette-id", speed=123.456)
    )
    assert subject.state.movement_speed_by_id["pipette-id"] == 123.456


def test_add_pipette_config(
    subject: PipetteStore,
    supported_tip_fixture: pipette_definition.SupportedTipsDefinition,
    available_sensors: pipette_definition.AvailableSensorDefinition,
) -> None:
    """It should update state from any pipette config private result."""
    dummy_command = create_comment_command()
    config = LoadedStaticPipetteData(
        model="pipette-model",
        display_name="pipette name",
        min_volume=1.23,
        max_volume=4.56,
        channels=7,
        flow_rates=FlowRates(
            default_aspirate={"a": 1},
            default_dispense={"b": 2},
            default_blow_out={"c": 3},
        ),
        tip_configuration_lookup_table={4: supported_tip_fixture},
        nominal_tip_overlap={"default": 5},
        home_position=8.9,
        nozzle_offset_z=10.11,
        nozzle_map=get_default_nozzle_map(PipetteNameType.P300_SINGLE),
        back_left_corner_offset=Point(x=1, y=2, z=3),
        front_right_corner_offset=Point(x=4, y=5, z=6),
        pipette_lld_settings={},
        plunger_positions={
            "top": 0.0,
            "bottom": 5.0,
            "blow_out": 19.0,
            "drop_tip": 20.0,
        },
        shaft_ul_per_mm=5.0,
        available_sensors=available_sensors,
        volume_mode=VolumeModes.default,
        available_volume_modes_min_vol={VolumeModes.default: 133.7},
    )

    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_config=update_types.PipetteConfigUpdate(
                    pipette_id="pipette-id",
                    config=config,
                    serial_number="pipette-serial",
                )
            ),
        )
    )

    assert subject.state.static_config_by_id["pipette-id"] == StaticPipetteConfig(
        model="pipette-model",
        serial_number="pipette-serial",
        display_name="pipette name",
        min_volume=1.23,
        max_volume=4.56,
        channels=7,
        tip_configuration_lookup_table={4: supported_tip_fixture},
        nominal_tip_overlap={"default": 5},
        home_position=8.9,
        nozzle_offset_z=10.11,
        bounding_nozzle_offsets=BoundingNozzlesOffsets(
            back_left_offset=Point(x=0, y=0, z=0),
            front_right_offset=Point(x=0, y=0, z=0),
        ),
        default_nozzle_map=get_default_nozzle_map(PipetteNameType.P300_SINGLE),
        pipette_bounding_box_offsets=PipetteBoundingBoxOffsets(
            back_left_corner=Point(x=1, y=2, z=3),
            front_right_corner=Point(x=4, y=5, z=6),
            front_left_corner=Point(x=1, y=5, z=3),
            back_right_corner=Point(x=4, y=2, z=3),
        ),
        lld_settings={},
        plunger_positions={
            "top": 0.0,
            "bottom": 5.0,
            "blow_out": 19.0,
            "drop_tip": 20.0,
        },
        shaft_ul_per_mm=5.0,
        available_sensors=available_sensors,
        volume_mode=VolumeModes.default,
        available_volume_modes_min_vol={VolumeModes.default: 133.7},
    )
    assert subject.state.flow_rates_by_id["pipette-id"].default_aspirate == {"a": 1.0}
    assert subject.state.flow_rates_by_id["pipette-id"].default_dispense == {"b": 2.0}
    assert subject.state.flow_rates_by_id["pipette-id"].default_blow_out == {"c": 3.0}


@pytest.mark.parametrize(
    "previous_state",
    [
        update_types.StateUpdate(
            pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                pipette_id="pipette-id"
            )
        ),
        update_types.StateUpdate(
            pipette_aspirated_fluid=update_types.PipetteEjectedFluidUpdate(
                pipette_id="pipette-id", volume=10
            )
        ),
    ],
)
def test_prepare_to_aspirate_marks_pipette_ready(
    subject: PipetteStore,
    previous_state: update_types.StateUpdate,
) -> None:
    """It should mark a pipette as ready to aspirate."""
    dummy_command = create_comment_command()
    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                loaded_pipette=update_types.LoadPipetteUpdate(
                    pipette_id="pipette-id",
                    pipette_name=PipetteNameType.P50_MULTI_FLEX,
                    mount=MountType.LEFT,
                    liquid_presence_detection=None,
                ),
                pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                    pipette_id="pipette-id"
                ),
            ),
        )
    )
    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_tip_state=update_types.PipetteTipStateUpdate(
                    pipette_id="pipette-id",
                    tip_geometry=TipGeometry(volume=42, length=101, diameter=8.0),
                    tip_source=LabwareWellId(labware_id="abc", well_name="123"),
                ),
                pipette_aspirated_fluid=update_types.PipetteEmptyFluidUpdate(
                    pipette_id="xyz", clean_tip=True
                ),
            ),
        )
    )

    subject.handle_action(
        SucceedCommandAction(command=dummy_command, state_update=previous_state)
    )

    subject.handle_action(
        SucceedCommandAction(
            command=dummy_command,
            state_update=update_types.StateUpdate(
                pipette_aspirated_fluid=update_types.PipetteEmptyFluidUpdate(
                    pipette_id="pipette-id", clean_tip=False
                )
            ),
        )
    )
    assert subject.state.pipette_contents_by_id["pipette-id"] == FluidStack()



"""Unit tests for the calibrateGripper implementation."""

from __future__ import annotations

import inspect
from datetime import datetime
from typing import TYPE_CHECKING

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import EarlyCapacitiveSenseTrigger

from opentrons.calibration_storage.types import (
    CalibrationStatus,
)
from opentrons.calibration_storage.types import (
    SourceType as CalibrationSourceType,
)
from opentrons.hardware_control import ot3_calibration
from opentrons.hardware_control.api import API as OT2API
from opentrons.hardware_control.instruments.ot3.instrument_calibration import (
    GripperCalibrationOffset,
)
from opentrons.hardware_control.types import GripperProbe, OT3Mount
from opentrons.protocol_engine.commands.calibration.calibrate_gripper import (
    CalibrateGripperImplementation,
    CalibrateGripperParams,
    CalibrateGripperParamsJaw,
    CalibrateGripperResult,
)
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.errors import HardwareNotSupportedError
from opentrons.protocol_engine.types import Vec3f
from opentrons.types import Point

if TYPE_CHECKING:
    # Support environments without OT-3 hardware control dependencies.
    from opentrons.hardware_control.ot3api import OT3API


@pytest.fixture(autouse=True)
def _mock_ot3_calibration(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    for name, func in inspect.getmembers(ot3_calibration, inspect.isfunction):
        monkeypatch.setattr(ot3_calibration, name, decoy.mock(func=func))


@pytest.mark.ot3_only
@pytest.mark.parametrize(
    "params_probe, expected_hc_probe",
    [
        (CalibrateGripperParamsJaw.FRONT, GripperProbe.FRONT),
        (CalibrateGripperParamsJaw.REAR, GripperProbe.REAR),
    ],
)
async def test_calibrate_gripper(
    decoy: Decoy,
    ot3_hardware_api: OT3API,
    _mock_ot3_calibration: None,
    params_probe: CalibrateGripperParamsJaw,
    expected_hc_probe: GripperProbe,
) -> None:
    """It should delegate to the hardware API to calibrate the gripper."""
    subject = CalibrateGripperImplementation(hardware_api=ot3_hardware_api)

    params = CalibrateGripperParams(jaw=params_probe)
    decoy.when(
        await ot3_calibration.calibrate_gripper_jaw(
            ot3_hardware_api, probe=expected_hc_probe
        )
    ).then_return(Point(1.1, 2.2, 3.3))

    result = await subject.execute(params)
    assert result == SuccessData(
        public=CalibrateGripperResult(jawOffset=Vec3f(x=1.1, y=2.2, z=3.3)),
    )


@pytest.mark.ot3_only
async def test_calibrate_gripper_saves_calibration(
    decoy: Decoy,
    ot3_hardware_api: OT3API,
    _mock_ot3_calibration: None,
) -> None:
    """It should delegate to hardware API to calibrate the gripper & save calibration."""
    subject = CalibrateGripperImplementation(hardware_api=ot3_hardware_api)
    params = CalibrateGripperParams(
        jaw=CalibrateGripperParamsJaw.REAR,
        otherJawOffset=Vec3f(x=4.4, y=5.5, z=6.6),
    )
    expected_calibration_data = GripperCalibrationOffset(
        offset=Point(x=101, y=102, z=103),
        source=CalibrationSourceType.calibration_check,
        status=CalibrationStatus(markedBad=False),
        last_modified=datetime(year=3000, month=1, day=1),
    )
    saved_delta_captor = matchers.Captor()
    decoy.when(
        await ot3_calibration.calibrate_gripper_jaw(
            ot3_hardware_api, probe=GripperProbe.REAR
        )
    ).then_return(Point(1.1, 2.2, 3.3))
    decoy.when(
        await ot3_hardware_api.save_instrument_offset(
            mount=OT3Mount.GRIPPER, delta=saved_delta_captor
        )
    ).then_return(expected_calibration_data)
    result = await subject.execute(params)
    saved_delta: Point = saved_delta_captor.value
    assert result.public.jawOffset == Vec3f(x=1.1, y=2.2, z=3.3)
    assert saved_delta.elementwise_isclose(Point(x=2.75, y=3.85, z=4.95))
    assert result.public.savedCalibration == expected_calibration_data


@pytest.mark.ot3_only
async def test_calibrate_gripper_does_not_save_during_error(
    decoy: Decoy, ot3_hardware_api: OT3API
) -> None:
    """Data should not be saved when an error is raised."""
    subject = CalibrateGripperImplementation(hardware_api=ot3_hardware_api)

    params = CalibrateGripperParams(
        jaw=CalibrateGripperParamsJaw.REAR,
        otherJawOffset=Vec3f(x=4.4, y=5.5, z=6.6),
    )

    decoy.when(
        await ot3_calibration.calibrate_gripper_jaw(
            ot3_hardware_api, probe=GripperProbe.REAR
        )
    ).then_raise(EarlyCapacitiveSenseTrigger(5.0, 3.0))

    with pytest.raises(EarlyCapacitiveSenseTrigger):
        await subject.execute(params)

    decoy.verify(
        await ot3_hardware_api.save_instrument_offset(
            mount=OT3Mount.LEFT, delta=Point(x=3, y=4, z=6)
        ),
        times=0,
    )


async def test_calibrate_gripper_raises_on_ot2(
    decoy: Decoy,
    ot2_hardware_api: OT2API,
) -> None:
    """It should raise with a descriptive error if run on an OT-2, instead of OT-3."""
    subject = CalibrateGripperImplementation(hardware_api=ot2_hardware_api)

    params = CalibrateGripperParams(jaw=CalibrateGripperParamsJaw.REAR)

    with pytest.raises(HardwareNotSupportedError):
        await subject.execute(params)



from typing import AsyncIterator, Dict

import pytest
from decoy import Decoy

from opentrons_hardware.drivers.can_bus import CanMessenger
from opentrons_hardware.firmware_bindings.constants import SensorId
from opentrons_hardware.hardware_control.tip_presence import (
    TipDetector,
)
from opentrons_hardware.hardware_control.tip_presence import (
    types as tp_types,
)
from opentrons_shared_data.errors.exceptions import UnmatchedTipPresenceStates

from opentrons.hardware_control.backends.tip_presence_manager import TipPresenceManager
from opentrons.hardware_control.types import InstrumentProbeType, OT3Mount, TipStateType


@pytest.fixture
def can_messenger(decoy: Decoy) -> CanMessenger:
    """Build a decoyed can messenger."""
    return decoy.mock(cls=CanMessenger)


@pytest.fixture
def tip_detector(decoy: Decoy) -> TipDetector:
    return decoy.mock(cls=TipDetector)


class TipDetectorController:
    def __init__(self, tip_detector: TipDetector, decoy: Decoy) -> None:
        self._tip_detector = tip_detector
        self._decoy = decoy

    async def retrieve_tip_status(self, tip_presence: bool) -> None:
        tip_notif = tp_types.TipNotification(sensor=SensorId.S0, presence=tip_presence)
        self._decoy.when(await self._tip_detector.request_tip_status()).then_return(
            [tip_notif]
        )

    async def retrieve_tip_status_highthroughput(
        self, tip_presences: Dict[SensorId, bool]
    ) -> None:
        tip_notif = [
            tp_types.TipNotification(sensor=sensor_id, presence=presence)
            for sensor_id, presence in tip_presences.items()
        ]
        self._decoy.when(await self._tip_detector.request_tip_status()).then_return(
            tip_notif
        )


@pytest.fixture
def tip_detector_controller(
    tip_detector: TipDetector,
    decoy: Decoy,
) -> TipDetectorController:
    return TipDetectorController(tip_detector, decoy)


@pytest.fixture
async def subject(
    can_messenger: CanMessenger,
    tip_detector: TipDetector,
) -> AsyncIterator[TipPresenceManager]:
    """Build a test subject using decoyed can messenger and tip detector."""
    manager = TipPresenceManager(can_messenger)
    manager.set_detector(OT3Mount.LEFT, tip_detector)
    try:
        yield manager
    finally:
        return


@pytest.mark.parametrize(
    "tip_presence,expected_type",
    [
        (False, TipStateType.ABSENT),
        (True, TipStateType.PRESENT),
    ],
)
async def test_get_tip_status_for_low_throughput(
    subject: TipPresenceManager,
    tip_detector_controller: TipDetectorController,
    tip_presence: bool,
    expected_type: TipStateType,
) -> None:
    mount = OT3Mount.LEFT
    await tip_detector_controller.retrieve_tip_status(tip_presence)

    result = await subject.get_tip_status(mount)
    result == expected_type


@pytest.mark.parametrize(
    "tip_presence,expected_type",
    [
        ({SensorId.S0: False, SensorId.S1: False}, TipStateType.ABSENT),
        ({SensorId.S0: True, SensorId.S1: True}, TipStateType.PRESENT),
    ],
)
async def test_get_tip_status_for_high_throughput(
    subject: TipPresenceManager,
    tip_detector_controller: TipDetectorController,
    tip_presence: Dict[SensorId, bool],
    expected_type: TipStateType,
) -> None:
    mount = OT3Mount.LEFT
    await tip_detector_controller.retrieve_tip_status_highthroughput(tip_presence)

    result = await subject.get_tip_status(mount)
    result == expected_type


@pytest.mark.parametrize(
    "tip_presence,expected_type,sensor_to_look_at",
    [
        (
            {SensorId.S0: False, SensorId.S1: False},
            TipStateType.ABSENT,
            InstrumentProbeType.PRIMARY,
        ),
        (
            {SensorId.S0: True, SensorId.S1: True},
            TipStateType.PRESENT,
            InstrumentProbeType.SECONDARY,
        ),
        (
            {SensorId.S0: False, SensorId.S1: True},
            TipStateType.ABSENT,
            InstrumentProbeType.PRIMARY,
        ),
        (
            {SensorId.S0: False, SensorId.S1: True},
            TipStateType.PRESENT,
            InstrumentProbeType.SECONDARY,
        ),
    ],
)
async def test_allow_different_tip_states_ht(
    subject: TipPresenceManager,
    tip_detector_controller: TipDetectorController,
    tip_presence: Dict[SensorId, bool],
    expected_type: TipStateType,
    sensor_to_look_at: InstrumentProbeType,
) -> None:
    mount = OT3Mount.LEFT
    await tip_detector_controller.retrieve_tip_status_highthroughput(tip_presence)

    result = await subject.get_tip_status(mount, sensor_to_look_at)
    result == expected_type

    # if sensor_to_look_at is not used, different tip states
    # should result in an UnmatchedTipStates error
    if len(set(tip_presence[t] for t in tip_presence)) > 1:
        with pytest.raises(UnmatchedTipPresenceStates):
            result = await subject.get_tip_status(mount)


@pytest.mark.parametrize(
    "tip_presence",
    [
        {SensorId.S0: True, SensorId.S1: False},
        {SensorId.S0: False, SensorId.S1: True},
    ],
)
async def test_unmatched_tip_responses_should_raise(
    subject: TipPresenceManager,
    tip_detector_controller: TipDetectorController,
    tip_presence: Dict[SensorId, bool],
) -> None:
    mount = OT3Mount.LEFT
    await tip_detector_controller.retrieve_tip_status_highthroughput(tip_presence)

    with pytest.raises(UnmatchedTipPresenceStates):
        await subject.get_tip_status(mount)



"""Command models to prepare the stacker shuttle for movement."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Union

from pydantic import BaseModel, Field
from typing_extensions import Type

from opentrons_shared_data.errors.exceptions import FlexStackerStallError

from ...errors import ErrorOccurrence
from ...resources import ModelUtils
from ..command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    DefinedErrorData,
    SuccessData,
)
from ..flex_stacker.common import FlexStackerStallOrCollisionError

if TYPE_CHECKING:
    from ...execution import EquipmentHandler
    from ...state.state import StateView

UnsafeFlexStackerPrepareShuttleCommandType = Literal[
    "unsafe/flexStacker/prepareShuttle"
]


class UnsafeFlexStackerPrepareShuttleParams(BaseModel):
    """The parameters for a UnsafeFlexStackerPrepareShuttle command."""

    moduleId: str = Field(..., description="Unique ID of the Flex Stacker")
    ignoreLatch: bool = Field(
        default=False, description="Ignore the latch state of the shuttle"
    )


class UnsafeFlexStackerPrepareShuttleResult(BaseModel):
    """Result data from a stacker UnsafeFlexStackerPrepareShuttle command."""


_ExecuteReturn = Union[
    SuccessData[UnsafeFlexStackerPrepareShuttleResult],
    DefinedErrorData[FlexStackerStallOrCollisionError],
]


class UnsafeFlexStackerPrepareShuttleImpl(
    AbstractCommandImpl[UnsafeFlexStackerPrepareShuttleParams, _ExecuteReturn]
):
    """Implementation of a stacker prepare shuttle command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        model_utils: ModelUtils,
        **kwargs: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment
        self._model_utils = model_utils

    async def execute(
        self, params: UnsafeFlexStackerPrepareShuttleParams
    ) -> _ExecuteReturn:
        """Execute the stacker prepare shuttle command.

        Moving the shuttle directly affects the state of the flex stacker and
        could affect its ability to execute the next command. This command
        should be used with care.
        """
        stacker_state = self._state_view.modules.get_flex_stacker_substate(
            params.moduleId
        )
        # Allow propagation of ModuleNotAttachedError.
        stacker_hw = self._equipment.get_module_hardware_api(stacker_state.module_id)

        try:
            if stacker_hw is not None:
                await stacker_hw.home_all(params.ignoreLatch)
        except FlexStackerStallError as e:
            return DefinedErrorData(
                public=FlexStackerStallOrCollisionError(
                    id=self._model_utils.generate_id(),
                    createdAt=self._model_utils.get_timestamp(),
                    wrappedErrors=[
                        ErrorOccurrence.from_failed(
                            id=self._model_utils.generate_id(),
                            createdAt=self._model_utils.get_timestamp(),
                            error=e,
                        )
                    ],
                    errorInfo={},
                ),
            )
        # TODO we should also add a check for shuttle not detected error

        return SuccessData(public=UnsafeFlexStackerPrepareShuttleResult())


class UnsafeFlexStackerPrepareShuttle(
    BaseCommand[
        UnsafeFlexStackerPrepareShuttleParams,
        UnsafeFlexStackerPrepareShuttleResult,
        ErrorOccurrence,
    ]
):
    """A command to prepare Flex Stacker shuttle."""

    commandType: UnsafeFlexStackerPrepareShuttleCommandType = (
        "unsafe/flexStacker/prepareShuttle"
    )
    params: UnsafeFlexStackerPrepareShuttleParams
    result: UnsafeFlexStackerPrepareShuttleResult | None = None

    _ImplementationCls: Type[UnsafeFlexStackerPrepareShuttleImpl] = (
        UnsafeFlexStackerPrepareShuttleImpl
    )


class UnsafeFlexStackerPrepareShuttleCreate(
    BaseCommandCreate[UnsafeFlexStackerPrepareShuttleParams]
):
    """A request to execute a Flex Stacker UnsafeFlexStackerPrepareShuttle command."""

    commandType: UnsafeFlexStackerPrepareShuttleCommandType = (
        "unsafe/flexStacker/prepareShuttle"
    )
    params: UnsafeFlexStackerPrepareShuttleParams

    _CommandCls: Type[UnsafeFlexStackerPrepareShuttle] = UnsafeFlexStackerPrepareShuttle



"""The versions of standard labware that the Protocol API should load by default."""

from typing import TypeAlias

from opentrons.protocols.api_support.types import APIVersion

DefaultLabwareVersions: TypeAlias = dict[APIVersion, dict[str, int]]


# This:
#
# {
#     APIVersion(2, 100): {
#         "foo_well_plate": 3,
#     },
#     APIVersion(2, 105): {
#         "foo_well_plate": 7
#     }
# }
#
# Means this:
#
# apiLevels        Load name         Default labware version
# ----------------------------------------------------------
# <2.100           foo_well_plate    1
# >=2.100,<2.105   foo_well_plate    3
# >=2.105          foo_well_plate    7
# [any]            [anything else]   1
DEFAULT_LABWARE_VERSIONS: DefaultLabwareVersions = {
    APIVersion(2, 14): {
        "armadillo_96_wellplate_200ul_pcr_full_skirt": 2,
        "biorad_384_wellplate_50ul": 2,
        "biorad_96_wellplate_200ul_pcr": 2,
        "corning_12_wellplate_6.9ml_flat": 2,
        "corning_24_wellplate_3.4ml_flat": 2,
        "corning_384_wellplate_112ul_flat": 2,
        "corning_48_wellplate_1.6ml_flat": 2,
        "corning_6_wellplate_16.8ml_flat": 2,
        "corning_96_wellplate_360ul_flat": 2,
        "nest_1_reservoir_195ml": 2,
        "nest_96_wellplate_100ul_pcr_full_skirt": 2,
        "nest_96_wellplate_200ul_flat": 2,
        "nest_96_wellplate_2ml_deep": 2,
        "opentrons_24_aluminumblock_generic_2ml_screwcap": 2,
        "opentrons_96_aluminumblock_generic_pcr_strip_200ul": 2,
        "opentrons_96_wellplate_200ul_pcr_full_skirt": 2,
    },
    APIVersion(2, 23): {
        "agilent_1_reservoir_290ml": 2,
        "appliedbiosystemsmicroamp_384_wellplate_40ul": 2,
        "armadillo_96_wellplate_200ul_pcr_full_skirt": 3,
        "axygen_1_reservoir_90ml": 2,
        "biorad_384_wellplate_50ul": 3,
        "biorad_96_wellplate_200ul_pcr": 3,
        "corning_12_wellplate_6.9ml_flat": 3,
        "corning_24_wellplate_3.4ml_flat": 3,
        "corning_384_wellplate_112ul_flat": 3,
        "corning_48_wellplate_1.6ml_flat": 3,
        "corning_6_wellplate_16.8ml_flat": 3,
        "corning_96_wellplate_360ul_flat": 3,
        "nest_12_reservoir_15ml": 2,
        "nest_1_reservoir_195ml": 3,
        "nest_1_reservoir_290ml": 2,
        "nest_96_wellplate_100ul_pcr_full_skirt": 3,
        "nest_96_wellplate_200ul_flat": 3,
        "nest_96_wellplate_2ml_deep": 3,
        "opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical": 2,
        "opentrons_10_tuberack_nest_4x50ml_6x15ml_conical": 2,
        "opentrons_15_tuberack_falcon_15ml_conical": 2,
        "opentrons_15_tuberack_nest_15ml_conical": 2,
        "opentrons_24_aluminumblock_generic_2ml_screwcap": 3,
        "opentrons_24_aluminumblock_nest_0.5ml_screwcap": 2,
        "opentrons_24_aluminumblock_nest_1.5ml_screwcap": 2,
        "opentrons_24_aluminumblock_nest_1.5ml_snapcap": 2,
        "opentrons_24_aluminumblock_nest_2ml_screwcap": 2,
        "opentrons_24_aluminumblock_nest_2ml_snapcap": 2,
        "opentrons_24_tuberack_eppendorf_1.5ml_safelock_snapcap": 2,
        "opentrons_24_tuberack_eppendorf_2ml_safelock_snapcap": 2,
        "opentrons_24_tuberack_generic_2ml_screwcap": 2,
        "opentrons_24_tuberack_nest_0.5ml_screwcap": 2,
        "opentrons_24_tuberack_nest_1.5ml_screwcap": 2,
        "opentrons_24_tuberack_nest_1.5ml_snapcap": 2,
        "opentrons_24_tuberack_nest_2ml_screwcap": 2,
        "opentrons_24_tuberack_nest_2ml_snapcap": 2,
        "opentrons_6_tuberack_falcon_50ml_conical": 2,
        "opentrons_6_tuberack_nest_50ml_conical": 2,
        "opentrons_96_aluminumblock_generic_pcr_strip_200ul": 3,
        "opentrons_96_wellplate_200ul_pcr_full_skirt": 3,
        "opentrons_tough_pcr_auto_sealing_lid": 2,
        "thermoscientificnunc_96_wellplate_1300ul": 2,
        "thermoscientificnunc_96_wellplate_2000ul": 2,
        "usascientific_12_reservoir_22ml": 2,
        "usascientific_96_wellplate_2.4ml_deep": 2,
    },
    APIVersion(2, 25): {
        "appliedbiosystemsmicroamp_384_wellplate_40ul": 3,
        "axygen_96_wellplate_500ul": 2,
        "biorad_384_wellplate_50ul": 4,
        "biorad_96_wellplate_200ul_pcr": 4,
        "corning_12_wellplate_6.9ml_flat": 4,
        "corning_24_wellplate_3.4ml_flat": 4,
        "corning_48_wellplate_1.6ml_flat": 5,
        "corning_6_wellplate_16.8ml_flat": 4,
        "corning_96_wellplate_360ul_flat": 4,
        "ibidi_96_square_well_plate_300ul": 2,
        "nest_96_wellplate_100ul_pcr_full_skirt": 4,
        "nest_96_wellplate_200ul_flat": 4,
        "nest_96_wellplate_2ml_deep": 4,
        "opentrons_96_wellplate_200ul_pcr_full_skirt": 4,
        "smc_384_read_plate": 2,
        "thermoscientificnunc_96_wellplate_1300ul": 3,
        "thermoscientificnunc_96_wellplate_2000ul": 3,
        "usascientific_96_wellplate_2.4ml_deep": 3,
    },
    APIVersion(2, 27): {
        "agilent_1_reservoir_290ml": 4,
        "axygen_1_reservoir_90ml": 3,
        "biorad_96_wellplate_200ul_pcr": 5,
        "corning_12_wellplate_6.9ml_flat": 5,
        "corning_24_wellplate_3.4ml_flat": 5,
        "corning_384_wellplate_112ul_flat": 5,
        "corning_48_wellplate_1.6ml_flat": 6,
        "corning_6_wellplate_16.8ml_flat": 5,
        "corning_96_wellplate_360ul_flat": 5,
        "nest_12_reservoir_15ml": 3,
        "nest_1_reservoir_195ml": 4,
        "nest_1_reservoir_290ml": 4,
        "nest_96_wellplate_100ul_pcr_full_skirt": 5,
        "nest_96_wellplate_200ul_flat": 5,
        "nest_96_wellplate_2ml_deep": 5,
        "opentrons_10_tuberack_falcon_4x50ml_6x15ml_conical": 3,
        "opentrons_15_tuberack_falcon_15ml_conical": 3,
        "opentrons_24_aluminumblock_nest_0.5ml_screwcap": 4,
        "opentrons_24_aluminumblock_nest_1.5ml_screwcap": 3,
        "opentrons_24_aluminumblock_nest_1.5ml_snapcap": 3,
        "opentrons_24_aluminumblock_nest_2ml_screwcap": 3,
        "opentrons_24_aluminumblock_nest_2ml_snapcap": 3,
        "opentrons_24_tuberack_eppendorf_1.5ml_safelock_snapcap": 3,
        "opentrons_24_tuberack_eppendorf_2ml_safelock_snapcap": 3,
        "opentrons_24_tuberack_nest_0.5ml_screwcap": 4,
        "opentrons_6_tuberack_nest_50ml_conical": 3,
        "opentrons_96_aluminumblock_generic_pcr_strip_200ul": 4,
        "opentrons_tough_universal_lid": 2,
        "usascientific_12_reservoir_22ml": 4,
        "usascientific_96_wellplate_2.4ml_deep": 4,
    },
    APIVersion(2, 28): {
        "black_96_well_microtiter_plate_lid": 2,
        "corning_96_wellplate_360ul_lid": 2,
        "corning_falcon_384_wellplate_130ul_flat_lid": 2,
    },
}


# Labware where, for whatever reason, we don't want `opentrons.protocol_api` to load
# the latest available version.
#
# Typically, this is because the latest available version of the labware is some kind of
# unpublicized draft.
#
# Beware, though, that users can still load the unpublicized draft if they know how, e.g.
# by passing an explicit `version` arg to `ProtocolContext.load_labware()`.
# And non-`opentrons.protocol_api` code like Labware Library, Protocol Designer, and
# Quick Transfer will still use the unpublicized draft unless you exclude it through
# other means.
#
# This list should not be consumed by production code--it's only for the benefit of tests
# that make sure every labware is accounted for somehow.
KNOWN_EXCEPTIONS_FOR_TESTS: set[str] = {
    # Dev testing junk for labware schema 3, not things that users should ever load:
    "schema3test_96_well_aluminum_block",
    "schema3test_96_wellplate_200ul_pcr_full_skirt",
    "schema3test_aluminum_flat_bottom_plate",
    "schema3test_flex_96_tiprack_200ul",
    "schema3test_flex_96_tiprack_adapter",
    "schema3test_flex_tiprack_lid",
    "schema3test_tough_pcr_auto_sealing_lid",
    "schema3test_universal_flat_adapter",
    "schema3test_96_wellplate_360ul_flat",
}


def get_standard_labware_default_version(
    api_version: APIVersion,
    load_name: str,
    default_labware_versions: DefaultLabwareVersions = DEFAULT_LABWARE_VERSIONS,
) -> int:
    """Return what version of a standard labware the Protocol API should load by default.

    The `default_labware_versions` param is exposed for testability and should be left
    unspecified.
    """
    default_labware_versions_newest_to_oldest = sorted(
        default_labware_versions.items(), key=lambda kv: kv[0], reverse=True
    )
    for (
        breakpoint_api_version,
        breakpoint_labware_versions,
    ) in default_labware_versions_newest_to_oldest:
        if (
            api_version >= breakpoint_api_version
            and load_name in breakpoint_labware_versions
        ):
            return breakpoint_labware_versions[load_name]

    return 1



"""Tests for the legacy Protocol API module core implementations."""

import pytest
from decoy import Decoy

from opentrons.hardware_control import SynchronousAdapter
from opentrons.hardware_control.modules import AbstractModule
from opentrons.hardware_control.modules.types import TemperatureModuleModel
from opentrons.protocol_api.core.legacy.legacy_module_core import LegacyModuleCore
from opentrons.protocol_api.core.legacy.legacy_protocol_core import (
    LegacyProtocolCore,
)
from opentrons.protocol_api.core.legacy.module_geometry import ModuleGeometry


@pytest.fixture
def mock_geometry(decoy: Decoy) -> ModuleGeometry:
    """Get a mock module geometry."""
    return decoy.mock(cls=ModuleGeometry)


@pytest.fixture
def mock_sync_module_hardware(decoy: Decoy) -> SynchronousAdapter[AbstractModule]:
    """Get a mock synchronous module hardware."""
    return decoy.mock(name="SynchronousAdapter[AbstractModule]")  # type: ignore[no-any-return]


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> LegacyProtocolCore:
    """Get a mock protocol core."""
    return decoy.mock(cls=LegacyProtocolCore)


@pytest.fixture
def subject(
    mock_geometry: ModuleGeometry,
    mock_sync_module_hardware: SynchronousAdapter[AbstractModule],
    mock_protocol_core: LegacyProtocolCore,
) -> LegacyModuleCore:
    """Get a legacy module implementation core with mocked out dependencies."""
    return LegacyModuleCore(
        requested_model=TemperatureModuleModel.TEMPERATURE_V1,
        geometry=mock_geometry,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )


def test_get_requested_model(subject: LegacyModuleCore) -> None:
    """It should return the requested model given to the constructor."""
    result = subject.get_requested_model()
    assert result == TemperatureModuleModel.TEMPERATURE_V1


def test_get_geometry(mock_geometry: ModuleGeometry, subject: LegacyModuleCore) -> None:
    """It should return the geometry interface given to the constructor."""
    assert subject.geometry is mock_geometry


def test_get_model(
    decoy: Decoy, mock_geometry: ModuleGeometry, subject: LegacyModuleCore
) -> None:
    """It should get the model from the geometry."""
    decoy.when(mock_geometry.model).then_return(TemperatureModuleModel.TEMPERATURE_V2)
    result = subject.get_model()
    assert result == "temperatureModuleV2"


def test_get_serial_number(
    decoy: Decoy,
    mock_sync_module_hardware: SynchronousAdapter[AbstractModule],
    subject: LegacyModuleCore,
) -> None:
    """It should return the serial number from the hardware interface."""
    decoy.when(mock_sync_module_hardware.device_info).then_return({"serial": "abc123"})
    result = subject.get_serial_number()
    assert result == "abc123"



import pytest
from decoy import Decoy

from opentrons_hardware.hardware_control.status_bar import StatusBar

from opentrons.hardware_control.backends.status_bar_state import (
    StatusBarStateController,
)
from opentrons.hardware_control.types import StatusBarState, StatusBarUpdateEvent


@pytest.fixture
def mock_status_bar_controller(decoy: Decoy) -> StatusBar:
    return decoy.mock(cls=StatusBar)


@pytest.fixture
def subject(mock_status_bar_controller: StatusBar) -> StatusBarStateController:
    return StatusBarStateController(mock_status_bar_controller)


SETTINGS = {
    StatusBarState.IDLE: StatusBarState.IDLE,
    StatusBarState.RUNNING: StatusBarState.RUNNING,
    StatusBarState.PAUSED: StatusBarState.PAUSED,
    StatusBarState.HARDWARE_ERROR: StatusBarState.HARDWARE_ERROR,
    StatusBarState.SOFTWARE_ERROR: StatusBarState.SOFTWARE_ERROR,
    StatusBarState.CONFIRMATION: StatusBarState.IDLE,
    StatusBarState.RUN_COMPLETED: StatusBarState.RUN_COMPLETED,
    StatusBarState.UPDATING: StatusBarState.UPDATING,
    StatusBarState.ACTIVATION: StatusBarState.IDLE,
    StatusBarState.DISCO: StatusBarState.IDLE,
    StatusBarState.OFF: StatusBarState.OFF,
}


@pytest.mark.parametrize(argnames=["enabled"], argvalues=[[True], [False]])
@pytest.mark.parametrize("setting", list(SETTINGS.keys()))
async def test_status_bar_interface(
    decoy: Decoy,
    subject: StatusBarStateController,
    enabled: bool,
    setting: StatusBarState,
) -> None:
    """Test setting status bar statuses and make sure the cached status is correct. Also
    verify that the listeners are called with the correct status bar state and enabled
    status."""
    listener_1 = decoy.mock(name="listener_1")
    listener_2 = decoy.mock(name="listener_2")
    subject.add_listener(listener_1)
    subject.add_listener(listener_2)

    await subject.set_enabled(enabled)
    await subject.set_status_bar_state(state=setting)
    assert subject.get_current_state() == SETTINGS[setting]

    decoy.verify(
        listener_1(StatusBarUpdateEvent(SETTINGS[setting], enabled)),
        listener_2(StatusBarUpdateEvent(SETTINGS[setting], enabled)),
    )



"""Test the Transfer class and its functions"""

from typing import TypedDict

import pytest

import opentrons.protocol_api as papi
from opentrons.hardware_control import ThreadManagedHardware
from opentrons.protocol_api.instrument_context import InstrumentContext
from opentrons.protocol_api.labware import Labware
from opentrons.protocol_api.protocol_context import ProtocolContext
from opentrons.protocols.advanced_control.common import MixStrategy
from opentrons.protocols.advanced_control.transfers import transfer as tx
from opentrons.protocols.api_support.types import APIVersion
from opentrons.types import Mount, TransferTipPolicy

# TODO (lc 12-8-2022) We need to re-write these transfer tests so that
# they are agnostic to the underlying hardware.
pytestmark = pytest.mark.ot2_only


class InstrLabware(TypedDict):
    ctx: ProtocolContext
    instr: InstrumentContext
    lw1: Labware
    lw2: Labware
    tiprack: Labware
    instr_multi: InstrumentContext


@pytest.fixture
def _instr_labware(ctx: ProtocolContext) -> InstrLabware:
    lw1 = ctx.load_labware("biorad_96_wellplate_200ul_pcr", 1)
    lw2 = ctx.load_labware("corning_96_wellplate_360ul_flat", 2)
    tiprack = ctx.load_labware("opentrons_96_tiprack_300ul", 3)
    tiprack2 = ctx.load_labware("opentrons_96_tiprack_300ul", 4)
    instr = ctx.load_instrument("p300_single", Mount.RIGHT, tip_racks=[tiprack])
    instr_multi = ctx.load_instrument("p300_multi", Mount.LEFT, tip_racks=[tiprack2])

    return {
        "ctx": ctx,
        "instr": instr,
        "lw1": lw1,
        "lw2": lw2,
        "tiprack": tiprack,
        "instr_multi": instr_multi,
    }


# +++++++ Test Helper Functions ++++++++++
def test_check_if_zero() -> None:
    tclass = tx.TransferPlan
    assert tclass._check_volume_not_zero(APIVersion(2, 6), 0)
    assert tclass._check_volume_not_zero(APIVersion(2, 6), 15)
    assert not tclass._check_volume_not_zero(APIVersion(2, 8), 0)
    assert tclass._check_volume_not_zero(APIVersion(2, 8), 15)


# +++++++ Test transfer types ++++++++++++
def test_default_transfers(_instr_labware: InstrLabware) -> None:
    # Transfer 100ml from row1 of labware1 to row1 of labware2: first with
    #  new_tip = ONCE, then with new_tip = NEVER
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]

    # ========== Transfer ===========
    xfer_plan = tx.TransferPlan(
        100,
        lw1.columns()[0],
        lw2.columns()[0],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="transfer",
    )
    xfer_plan_list = []
    for step in xfer_plan:
        xfer_plan_list.append(step)
    exp1 = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][0], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[0][0], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][1], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[0][1], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][2], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[0][2], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][3], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[0][3], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][4], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[0][4], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][5], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[0][5], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][6], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[0][6], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][7], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[0][7], 1.0], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    assert xfer_plan_list == exp1

    # ========== Distribute ===========
    dist_plan = tx.TransferPlan(
        50,
        lw1.columns()[0][0],
        lw2.columns()[0],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="distribute",
    )
    dist_plan_list = []
    for step in dist_plan:
        dist_plan_list.append(step)
    exp2 = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [300, lw1.columns()[0][0], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [50, lw2.columns()[0][0], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [50, lw2.columns()[0][1], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [50, lw2.columns()[0][2], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [50, lw2.columns()[0][3], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [50, lw2.columns()[0][4], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [50, lw2.columns()[0][5], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][0], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [50, lw2.columns()[0][6], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [50, lw2.columns()[0][7], 1.0], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    assert dist_plan_list == exp2

    # ========== Consolidate ===========
    consd_plan = tx.TransferPlan(
        50,
        lw1.columns()[0],
        lw2.columns()[0][0],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="consolidate",
    )
    consd_plan_list = []
    for step in consd_plan:
        consd_plan_list.append(step)
    exp3 = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [50, lw1.columns()[0][0], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [50, lw1.columns()[0][1], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [50, lw1.columns()[0][2], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [50, lw1.columns()[0][3], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [50, lw1.columns()[0][4], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [50, lw1.columns()[0][5], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [300, lw2.columns()[0][0], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [50, lw1.columns()[0][6], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [50, lw1.columns()[0][7], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[0][0], 1.0], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    assert consd_plan_list == exp3


def test_uneven_transfers(_instr_labware: InstrLabware) -> None:
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]

    options = tx.TransferOptions()
    options = options._replace(
        transfer=options.transfer._replace(new_tip=TransferTipPolicy.NEVER)
    )

    # ========== One-to-Many ==========
    xfer_plan = tx.TransferPlan(
        100,
        lw1.columns()[0][0],
        lw2.columns()[1][:4],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="transfer",
        options=options,
    )
    one_to_many_plan_list = []
    for step in xfer_plan:
        one_to_many_plan_list.append(step)
    exp1 = [
        {"method": "aspirate", "args": [100, lw1.columns()[0][0], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[1][0], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][0], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[1][1], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][0], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[1][2], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][0], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[1][3], 1.0], "kwargs": {}},
    ]
    assert one_to_many_plan_list == exp1

    # ========== Few-to-Many ==========
    xfer_plan = tx.TransferPlan(
        [100, 90, 80, 70],
        lw1.columns()[0][:2],
        lw2.columns()[1][:4],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="transfer",
        options=options,
    )
    few_to_many_plan_list = []
    for step in xfer_plan:
        few_to_many_plan_list.append(step)
    exp2 = [
        {"method": "aspirate", "args": [100, lw1.columns()[0][0], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[1][0], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [90, lw1.columns()[0][0], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [90, lw2.columns()[1][1], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [80, lw1.columns()[0][1], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [80, lw2.columns()[1][2], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [70, lw1.columns()[0][1], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [70, lw2.columns()[1][3], 1.0], "kwargs": {}},
    ]
    assert few_to_many_plan_list == exp2

    # ========== Many-to-One ==========
    xfer_plan = tx.TransferPlan(
        [100, 90, 80, 70],
        lw1.columns()[0][:4],
        lw2.columns()[1][0],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="transfer",
        options=options,
    )
    many_to_one_plan_list = []
    for step in xfer_plan:
        many_to_one_plan_list.append(step)
    exp3 = [
        {"method": "aspirate", "args": [100, lw1.columns()[0][0], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[1][0], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [90, lw1.columns()[0][1], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [90, lw2.columns()[1][0], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [80, lw1.columns()[0][2], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [80, lw2.columns()[1][0], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [70, lw1.columns()[0][3], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [70, lw2.columns()[1][0], 1.0], "kwargs": {}},
    ]
    assert many_to_one_plan_list == exp3


def test_location_wells(_instr_labware: InstrLabware) -> None:
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]
    aspirate_loc = lw1.wells()[0].top()
    # Test single-channel transfer with locations
    list_of_locs = [well.bottom(5) for col in lw2.columns()[0:11] for well in col]

    xfer_plan = tx.TransferPlan(
        30,
        aspirate_loc,
        list_of_locs,
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="transfer",
    )
    idx_dest = 0
    for step in xfer_plan:
        if step["method"] == "aspirate":
            assert isinstance(step["args"], list)
            assert step["args"][1].point == aspirate_loc.point
        elif step["method"] == "dispense":
            assert isinstance(step["args"], list)
            assert step["args"][1].point == list_of_locs[idx_dest].point
            idx_dest += 1

    multi_locs = [col[0].bottom(5) for col in lw2.columns()[0:11]]
    # Test multi-channel transfer with locations
    xfer_plan = tx.TransferPlan(
        30,
        aspirate_loc,
        multi_locs,
        _instr_labware["instr_multi"],
        max_volume=_instr_labware["instr_multi"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="transfer",
    )

    idx_dest = 0
    for step in xfer_plan:
        if step["method"] == "aspirate":
            assert isinstance(step["args"], list)
            assert step["args"][1].point == aspirate_loc.point
        elif step["method"] == "dispense":
            assert isinstance(step["args"], list)
            assert step["args"][1].point == multi_locs[idx_dest].point
            idx_dest += 1


def test_no_new_tip(_instr_labware: InstrLabware) -> None:
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]

    options = tx.TransferOptions()
    options = options._replace(
        transfer=options.transfer._replace(new_tip=TransferTipPolicy.NEVER)
    )
    # ========== Transfer ==========
    xfer_plan = tx.TransferPlan(
        100,
        lw1.columns()[0],
        lw2.columns()[0],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="transfer",
        options=options,
    )
    for step in xfer_plan:
        assert step["method"] != "pick_up_tip"
        assert step["method"] != "drop_tip"

    # ========== Distribute ===========
    dist_plan = tx.TransferPlan(
        30,
        lw1.columns()[0][0],
        lw2.columns()[0],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="distribute",
        options=options,
    )
    for step in dist_plan:
        assert step["method"] != "pick_up_tip"
        assert step["method"] != "drop_tip"

    # ========== Consolidate ===========
    consd_plan = tx.TransferPlan(
        40,
        lw1.columns()[0],
        lw2.rows()[0][1],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="transfer",
        options=options,
    )
    for step in consd_plan:
        assert step["method"] != "pick_up_tip"
        assert step["method"] != "drop_tip"


def test_new_tip_always(
    _instr_labware: InstrLabware, monkeypatch: pytest.MonkeyPatch
) -> None:
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]
    tiprack = _instr_labware["tiprack"]
    i_ctx = _instr_labware["instr"]

    options = tx.TransferOptions()
    options = options._replace(
        transfer=options.transfer._replace(
            new_tip=TransferTipPolicy.ALWAYS, drop_tip_strategy=tx.DropTipStrategy.TRASH
        )
    )

    xfer_plan = tx.TransferPlan(
        100,
        lw1.columns()[0][1:5],
        lw2.columns()[0][1:5],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="transfer",
        options=options,
    )
    xfer_plan_list = []
    for step in xfer_plan:
        xfer_plan_list.append(step)
    exp1 = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][1], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[0][1], 1.0], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][2], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[0][2], 1.0], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][3], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[0][3], 1.0], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][4], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.columns()[0][4], 1.0], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    assert xfer_plan_list == exp1
    for cmd in xfer_plan_list:
        getattr(i_ctx, cmd["method"])(*cmd["args"], **cmd["kwargs"])
    assert tiprack.next_tip() == tiprack.columns()[0][4]


def test_transfer_w_touchtip_blowout(_instr_labware: InstrLabware) -> None:
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]
    assert hasattr(_instr_labware["instr"].trash_container, "wells")

    # ========== Transfer ==========
    options = tx.TransferOptions()
    options = options._replace(
        transfer=options.transfer._replace(
            touch_tip_strategy=tx.TouchTipStrategy.ALWAYS,
            blow_out_strategy=tx.BlowOutStrategy.TRASH,
            new_tip=TransferTipPolicy.NEVER,
        )
    )

    xfer_plan = tx.TransferPlan(
        100,
        lw1.columns()[0][:3],
        lw2.rows()[0][:3],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="transfer",
        options=options,
    )
    xfer_plan_list = []
    for step in xfer_plan:
        xfer_plan_list.append(step)
    exp1 = [
        {"method": "aspirate", "args": [100, lw1.columns()[0][0], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.rows()[0][0], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {
            "method": "blow_out",
            "args": [_instr_labware["instr"].trash_container.wells()[0]],
            "kwargs": {},
        },
        {"method": "aspirate", "args": [100, lw1.columns()[0][1], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.rows()[0][1], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {
            "method": "blow_out",
            "args": [_instr_labware["instr"].trash_container.wells()[0]],
            "kwargs": {},
        },
        {"method": "aspirate", "args": [100, lw1.columns()[0][2], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.rows()[0][2], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {
            "method": "blow_out",
            "args": [_instr_labware["instr"].trash_container.wells()[0]],
            "kwargs": {},
        },
    ]
    assert xfer_plan_list == exp1

    # ========== Distribute ==========
    options = tx.TransferOptions()
    options = options._replace(
        transfer=options.transfer._replace(
            disposal_volume=_instr_labware["instr"].min_volume,
            touch_tip_strategy=tx.TouchTipStrategy.ALWAYS,
            new_tip=TransferTipPolicy.NEVER,
        )
    )

    dist_plan = tx.TransferPlan(
        30,
        lw1.columns()[0][0],
        lw2.rows()[0][:3],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="distribute",
        options=options,
    )
    dist_plan_list = []
    for step in dist_plan:
        dist_plan_list.append(step)
    exp2 = [
        {"method": "aspirate", "args": [120, lw1.columns()[0][0], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "dispense", "args": [30, lw2.rows()[0][0], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "dispense", "args": [30, lw2.rows()[0][1], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "dispense", "args": [30, lw2.rows()[0][2], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {
            "method": "blow_out",
            "args": [_instr_labware["instr"].trash_container.wells()[0]],
            "kwargs": {},
        },
    ]
    assert dist_plan_list == exp2


def test_transfer_w_airgap_blowout(_instr_labware: InstrLabware) -> None:
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]

    options = tx.TransferOptions()
    options = options._replace(
        transfer=options.transfer._replace(
            air_gap=10,
            blow_out_strategy=tx.BlowOutStrategy.DEST,
            new_tip=TransferTipPolicy.NEVER,
        )
    )

    # ========== Transfer ==========
    xfer_plan = tx.TransferPlan(
        100,
        lw1.columns()[0][1:5],
        lw2.rows()[0][1:5],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="transfer",
        options=options,
    )
    xfer_plan_list = []
    for step in xfer_plan:
        xfer_plan_list.append(step)
    exp1 = [
        {"method": "aspirate", "args": [100, lw1.columns()[0][1], 1.0], "kwargs": {}},
        {"method": "air_gap", "args": [10], "kwargs": {}},
        {"method": "dispense", "args": [110, lw2.rows()[0][1], 1.0], "kwargs": {}},
        {"method": "blow_out", "args": [lw2.rows()[0][1]], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][2], 1.0], "kwargs": {}},
        {"method": "air_gap", "args": [10], "kwargs": {}},
        {"method": "dispense", "args": [110, lw2.rows()[0][2], 1.0], "kwargs": {}},
        {"method": "blow_out", "args": [lw2.rows()[0][2]], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][3], 1.0], "kwargs": {}},
        {"method": "air_gap", "args": [10], "kwargs": {}},
        {"method": "dispense", "args": [110, lw2.rows()[0][3], 1.0], "kwargs": {}},
        {"method": "blow_out", "args": [lw2.rows()[0][3]], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][4], 1.0], "kwargs": {}},
        {"method": "air_gap", "args": [10], "kwargs": {}},
        {"method": "dispense", "args": [110, lw2.rows()[0][4], 1.0], "kwargs": {}},
        {"method": "blow_out", "args": [lw2.rows()[0][4]], "kwargs": {}},
    ]
    assert xfer_plan_list == exp1

    # ========== Distribute ==========
    dist_plan = tx.TransferPlan(
        60,
        lw1.columns()[1][0],
        lw2.rows()[1][1:6],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="distribute",
        options=options,
    )
    dist_plan_list = []
    for step in dist_plan:
        dist_plan_list.append(step)
    exp2 = [
        {"method": "aspirate", "args": [240, lw1.columns()[1][0], 1.0], "kwargs": {}},
        {"method": "air_gap", "args": [10], "kwargs": {}},
        {"method": "dispense", "args": [70, lw2.rows()[1][1], 1.0], "kwargs": {}},
        {"method": "air_gap", "args": [10], "kwargs": {}},
        {"method": "dispense", "args": [70, lw2.rows()[1][2], 1.0], "kwargs": {}},
        {"method": "air_gap", "args": [10], "kwargs": {}},
        {"method": "dispense", "args": [70, lw2.rows()[1][3], 1.0], "kwargs": {}},
        {"method": "air_gap", "args": [10], "kwargs": {}},
        {"method": "dispense", "args": [70, lw2.rows()[1][4], 1.0], "kwargs": {}},
        {"method": "blow_out", "args": [lw2.rows()[1][4]], "kwargs": {}},
        {"method": "aspirate", "args": [60, lw1.columns()[1][0], 1.0], "kwargs": {}},
        {"method": "air_gap", "args": [10], "kwargs": {}},
        {"method": "dispense", "args": [70, lw2.rows()[1][5], 1.0], "kwargs": {}},
        {"method": "blow_out", "args": [lw2.rows()[1][5]], "kwargs": {}},
    ]
    assert dist_plan_list == exp2

    # ========== Consolidate ==========
    consd_plan = tx.TransferPlan(
        60,
        lw1.columns()[1],
        lw2.rows()[1][1],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="consolidate",
        options=options,
    )
    consd_plan_list = []
    for step in consd_plan:
        consd_plan_list.append(step)
    exp3 = [
        {"method": "aspirate", "args": [60, lw1.columns()[1][0], 1.0], "kwargs": {}},
        {"method": "air_gap", "args": [10], "kwargs": {}},
        {"method": "aspirate", "args": [60, lw1.columns()[1][1], 1.0], "kwargs": {}},
        {"method": "air_gap", "args": [10], "kwargs": {}},
        {"method": "aspirate", "args": [60, lw1.columns()[1][2], 1.0], "kwargs": {}},
        {"method": "air_gap", "args": [10], "kwargs": {}},
        {"method": "aspirate", "args": [60, lw1.columns()[1][3], 1.0], "kwargs": {}},
        {"method": "air_gap", "args": [10], "kwargs": {}},
        {"method": "dispense", "args": [280, lw2.rows()[1][1], 1.0], "kwargs": {}},
        {"method": "blow_out", "args": [lw2.rows()[1][1]], "kwargs": {}},
        {"method": "aspirate", "args": [60, lw1.columns()[1][4], 1.0], "kwargs": {}},
        {"method": "air_gap", "args": [10], "kwargs": {}},
        {"method": "aspirate", "args": [60, lw1.columns()[1][5], 1.0], "kwargs": {}},
        {"method": "air_gap", "args": [10], "kwargs": {}},
        {"method": "aspirate", "args": [60, lw1.columns()[1][6], 1.0], "kwargs": {}},
        {"method": "air_gap", "args": [10], "kwargs": {}},
        {"method": "aspirate", "args": [60, lw1.columns()[1][7], 1.0], "kwargs": {}},
        {"method": "air_gap", "args": [10], "kwargs": {}},
        {"method": "dispense", "args": [280, lw2.rows()[1][1], 1.0], "kwargs": {}},
        {"method": "blow_out", "args": [lw2.rows()[1][1]], "kwargs": {}},
    ]
    assert consd_plan_list == exp3


def test_touchtip_mix(_instr_labware: InstrLabware) -> None:
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]

    options = tx.TransferOptions()
    options = options._replace(
        transfer=options.transfer._replace(
            new_tip=TransferTipPolicy.NEVER,
            touch_tip_strategy=tx.TouchTipStrategy.ALWAYS,
            mix_strategy=MixStrategy.AFTER,
        )
    )

    # ========== Transfer ==========
    xfer_plan = tx.TransferPlan(
        100,
        lw1.columns()[0][1:5],
        lw2.rows()[0][1:5],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="transfer",
        options=options,
    )
    xfer_plan_list = []
    for step in xfer_plan:
        xfer_plan_list.append(step)
    exp1 = [
        {"method": "aspirate", "args": [100, lw1.columns()[0][1], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.rows()[0][1], 1.0], "kwargs": {}},
        {"method": "mix", "args": [], "kwargs": {"location": lw2.rows()[0][1]}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][2], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.rows()[0][2], 1.0], "kwargs": {}},
        {"method": "mix", "args": [], "kwargs": {"location": lw2.rows()[0][2]}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][3], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.rows()[0][3], 1.0], "kwargs": {}},
        {"method": "mix", "args": [], "kwargs": {"location": lw2.rows()[0][3]}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][4], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2.rows()[0][4], 1.0], "kwargs": {}},
        {"method": "mix", "args": [], "kwargs": {"location": lw2.rows()[0][4]}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
    ]
    assert xfer_plan_list == exp1

    # ========== Distribute ==========
    dist_plan = tx.TransferPlan(
        60,
        lw1.columns()[1][0],
        lw2.rows()[1][1:6],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="distribute",
        options=options,
    )
    dist_plan_list = []
    for step in dist_plan:
        dist_plan_list.append(step)
    exp2 = [
        {"method": "aspirate", "args": [300, lw1.columns()[1][0], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "dispense", "args": [60, lw2.rows()[1][1], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "dispense", "args": [60, lw2.rows()[1][2], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "dispense", "args": [60, lw2.rows()[1][3], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "dispense", "args": [60, lw2.rows()[1][4], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "dispense", "args": [60, lw2.rows()[1][5], 1.0], "kwargs": {}},
        {"method": "mix", "args": [], "kwargs": {"location": lw2.rows()[1][5]}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
    ]

    assert dist_plan_list == exp2

    # ========== Consolidate ==========
    consd_plan = tx.TransferPlan(
        60,
        lw1.columns()[1],
        lw2.rows()[1][1],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="consolidate",
        options=options,
    )
    consd_plan_list = []
    for step in consd_plan:
        consd_plan_list.append(step)
    exp3 = [
        {"method": "aspirate", "args": [60, lw1.columns()[1][0], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [60, lw1.columns()[1][1], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [60, lw1.columns()[1][2], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [60, lw1.columns()[1][3], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [60, lw1.columns()[1][4], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "dispense", "args": [300, lw2.rows()[1][1], 1.0], "kwargs": {}},
        {"method": "mix", "args": [], "kwargs": {"location": lw2.rows()[1][1]}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [60, lw1.columns()[1][5], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [60, lw1.columns()[1][6], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [60, lw1.columns()[1][7], 1.0], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
        {"method": "dispense", "args": [180, lw2.rows()[1][1], 1.0], "kwargs": {}},
        {"method": "mix", "args": [], "kwargs": {"location": lw2.rows()[1][1]}},
        {"method": "touch_tip", "args": [], "kwargs": {}},
    ]
    assert consd_plan_list == exp3


def test_all_options(_instr_labware: InstrLabware) -> None:
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]

    options = tx.TransferOptions()
    options = options._replace(
        transfer=options.transfer._replace(
            new_tip=TransferTipPolicy.ONCE,
            drop_tip_strategy=tx.DropTipStrategy.RETURN,
            touch_tip_strategy=tx.TouchTipStrategy.ALWAYS,
            mix_strategy=MixStrategy.AFTER,
        ),
        pick_up_tip=options.pick_up_tip._replace(presses=4, increment=2),
        touch_tip=options.touch_tip._replace(speed=1.6),
        mix=options.mix._replace(
            mix_after=options.mix.mix_after._replace(repetitions=2)
        ),
        blow_out=options.blow_out._replace(location=lw2.columns()[10][0]),
        aspirate=options.aspirate._replace(rate=1.5),
    )

    xfer_plan = tx.TransferPlan(
        100,
        lw1.columns()[0][1:4],
        lw2.rows()[0][1:4],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="transfer",
        options=options,
    )
    xfer_plan_list = []
    for step in xfer_plan:
        xfer_plan_list.append(step)
    exp1 = [
        {"method": "pick_up_tip", "args": [], "kwargs": {"presses": 4, "increment": 2}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][1], 1.5], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {"speed": 1.6}},
        {"method": "dispense", "args": [100, lw2.rows()[0][1], 1.0], "kwargs": {}},
        {
            "method": "mix",
            "args": [],
            "kwargs": {"repetitions": 2, "location": lw2.rows()[0][1]},
        },
        {"method": "touch_tip", "args": [], "kwargs": {"speed": 1.6}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][2], 1.5], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {"speed": 1.6}},
        {"method": "dispense", "args": [100, lw2.rows()[0][2], 1.0], "kwargs": {}},
        {
            "method": "mix",
            "args": [],
            "kwargs": {"repetitions": 2, "location": lw2.rows()[0][2]},
        },
        {"method": "touch_tip", "args": [], "kwargs": {"speed": 1.6}},
        {"method": "aspirate", "args": [100, lw1.columns()[0][3], 1.5], "kwargs": {}},
        {"method": "touch_tip", "args": [], "kwargs": {"speed": 1.6}},
        {"method": "dispense", "args": [100, lw2.rows()[0][3], 1.0], "kwargs": {}},
        {
            "method": "mix",
            "args": [],
            "kwargs": {"repetitions": 2, "location": lw2.rows()[0][3]},
        },
        {"method": "touch_tip", "args": [], "kwargs": {"speed": 1.6}},
        {"method": "return_tip", "args": [], "kwargs": {}},
    ]
    assert xfer_plan_list == exp1


def test_invalid_air_gap_disposal_sum_distribute(_instr_labware: InstrLabware) -> None:
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]
    # Supply the disposal volume assessment with the instrument followed by a disposal value equal to the max volume

    options = tx.TransferOptions(
        transfer=tx.Transfer(
            disposal_volume=_instr_labware["instr"].max_volume / 2,
            air_gap=_instr_labware["instr"].max_volume / 2,
        )
    )
    with pytest.raises(ValueError):
        plan = tx.TransferPlan(
            700,
            lw1.columns()[0][0],
            lw2.rows()[0][1:3],
            _instr_labware["instr"],
            max_volume=_instr_labware["instr"].hw_pipette["max_volume"],
            api_version=_instr_labware["ctx"].api_version,
            mode="distribute",
            options=options,
        )
        # Exhaust the iterator in case it raises the expected exception lazily.
        list(plan)


def test_invalid_air_gap_distribute(_instr_labware: InstrLabware) -> None:
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]
    # Supply the disposal volume assessment with the instrument followed by a disposal value equal to the max volume

    options = tx.TransferOptions(
        transfer=tx.Transfer(air_gap=_instr_labware["instr"].max_volume)
    )
    with pytest.raises(ValueError):
        plan = tx.TransferPlan(
            700,
            lw1.columns()[0][0],
            lw2.rows()[0][1:3],
            _instr_labware["instr"],
            max_volume=_instr_labware["instr"].hw_pipette["max_volume"],
            api_version=_instr_labware["ctx"].api_version,
            mode="distribute",
            options=options,
        )
        # Exhaust the iterator in case it raises the expected exception lazily.
        list(plan)


def test_invalid_disposal_volume_distribute(_instr_labware: InstrLabware) -> None:
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]
    # Supply the disposal volume assessment with the instrument followed by a disposal value equal to the max volume

    options = tx.TransferOptions(
        transfer=tx.Transfer(disposal_volume=_instr_labware["instr"].max_volume)
    )
    with pytest.raises(ValueError):
        plan = tx.TransferPlan(
            700,
            lw1.columns()[0][0],
            lw2.rows()[0][1:3],
            _instr_labware["instr"],
            max_volume=_instr_labware["instr"].hw_pipette["max_volume"],
            api_version=_instr_labware["ctx"].api_version,
            mode="distribute",
            options=options,
        )
        # Exhaust the iterator in case it raises the expected exception lazily.
        list(plan)


def test_oversized_distribute(_instr_labware: InstrLabware) -> None:
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]

    xfer_plan = tx.TransferPlan(
        700,
        lw1.columns()[0][0],
        lw2.rows()[0][1:3],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="distribute",
    )
    xfer_plan_list = []
    for step in xfer_plan:
        xfer_plan_list.append(step)
    exp1 = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {
            "method": "aspirate",
            "args": [300, lw1.wells_by_index()["A1"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [300, lw2.wells_by_index()["A2"], 1.0],
            "kwargs": {},
        },
        {
            "method": "aspirate",
            "args": [200, lw1.wells_by_index()["A1"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [200, lw2.wells_by_index()["A2"], 1.0],
            "kwargs": {},
        },
        {
            "method": "aspirate",
            "args": [200, lw1.wells_by_index()["A1"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [200, lw2.wells_by_index()["A2"], 1.0],
            "kwargs": {},
        },
        {
            "method": "aspirate",
            "args": [300, lw1.wells_by_index()["A1"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [300, lw2.wells_by_index()["A3"], 1.0],
            "kwargs": {},
        },
        {
            "method": "aspirate",
            "args": [200, lw1.wells_by_index()["A1"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [200, lw2.wells_by_index()["A3"], 1.0],
            "kwargs": {},
        },
        {
            "method": "aspirate",
            "args": [200, lw1.wells_by_index()["A1"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [200, lw2.wells_by_index()["A3"], 1.0],
            "kwargs": {},
        },
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    assert xfer_plan_list == exp1


def test_oversized_consolidate(_instr_labware: InstrLabware) -> None:
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]

    xfer_plan = tx.TransferPlan(
        700,
        lw2.rows()[0][1:3],
        lw1.wells_by_index()["A1"],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="consolidate",
    )
    xfer_plan_list = []
    for step in xfer_plan:
        xfer_plan_list.append(step)
    exp1 = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {
            "method": "aspirate",
            "args": [300, lw2.wells_by_index()["A2"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [300, lw1.wells_by_index()["A1"], 1.0],
            "kwargs": {},
        },
        {
            "method": "aspirate",
            "args": [200, lw2.wells_by_index()["A2"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [200, lw1.wells_by_index()["A1"], 1.0],
            "kwargs": {},
        },
        {
            "method": "aspirate",
            "args": [200, lw2.wells_by_index()["A2"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [200, lw1.wells_by_index()["A1"], 1.0],
            "kwargs": {},
        },
        {
            "method": "aspirate",
            "args": [300, lw2.wells_by_index()["A3"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [300, lw1.wells_by_index()["A1"], 1.0],
            "kwargs": {},
        },
        {
            "method": "aspirate",
            "args": [200, lw2.wells_by_index()["A3"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [200, lw1.wells_by_index()["A1"], 1.0],
            "kwargs": {},
        },
        {
            "method": "aspirate",
            "args": [200, lw2.wells_by_index()["A3"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [200, lw1.wells_by_index()["A1"], 1.0],
            "kwargs": {},
        },
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    assert xfer_plan_list == exp1


def test_oversized_transfer(_instr_labware: InstrLabware) -> None:
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]

    xfer_plan = tx.TransferPlan(
        700,
        lw2.rows()[0][1:3],
        lw1.columns()[0][1:3],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=_instr_labware["ctx"].api_version,
        mode="transfer",
    )
    xfer_plan_list = []
    for step in xfer_plan:
        xfer_plan_list.append(step)
    exp1 = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {
            "method": "aspirate",
            "args": [300, lw2.wells_by_index()["A2"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [300, lw1.wells_by_index()["B1"], 1.0],
            "kwargs": {},
        },
        {
            "method": "aspirate",
            "args": [200, lw2.wells_by_index()["A2"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [200, lw1.wells_by_index()["B1"], 1.0],
            "kwargs": {},
        },
        {
            "method": "aspirate",
            "args": [200, lw2.wells_by_index()["A2"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [200, lw1.wells_by_index()["B1"], 1.0],
            "kwargs": {},
        },
        {
            "method": "aspirate",
            "args": [300, lw2.wells_by_index()["A3"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [300, lw1.wells_by_index()["C1"], 1.0],
            "kwargs": {},
        },
        {
            "method": "aspirate",
            "args": [200, lw2.wells_by_index()["A3"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [200, lw1.wells_by_index()["C1"], 1.0],
            "kwargs": {},
        },
        {
            "method": "aspirate",
            "args": [200, lw2.wells_by_index()["A3"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [200, lw1.wells_by_index()["C1"], 1.0],
            "kwargs": {},
        },
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    assert xfer_plan_list == exp1


def test_multichannel_transfer_old_version(
    hardware: ThreadManagedHardware, deck_definition_name: str
) -> None:
    # for API version below 2.2, multichannel pipette can only
    # reach row A of 384-well plates
    ctx = papi.create_protocol_context(
        api_version=APIVersion(2, 1),
        hardware_api=hardware,
        deck_type=deck_definition_name,
    )

    lw1 = ctx.load_labware("biorad_96_wellplate_200ul_pcr", 1)
    lw2 = ctx.load_labware("corning_384_wellplate_112ul_flat", 2)
    tiprack = ctx.load_labware("opentrons_96_tiprack_300ul", 3)
    instr_multi = ctx.load_instrument("p300_multi", Mount.LEFT, tip_racks=[tiprack])

    xfer_plan = tx.TransferPlan(
        100,
        lw1.rows()[0][0],
        [lw2.rows()[0][1], lw2.rows()[1][1]],
        instr_multi,
        max_volume=instr_multi.hw_pipette["working_volume"],
        api_version=ctx.api_version,
        mode="distribute",
    )
    xfer_plan_list = []
    for step in xfer_plan:
        xfer_plan_list.append(step)
    exp1 = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {
            "method": "aspirate",
            "args": [100, lw1.wells_by_name()["A1"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [100, lw2.wells_by_index()["A2"], 1.0],
            "kwargs": {},
        },
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    assert xfer_plan_list == exp1

    # target without row limit
    with pytest.raises(IndexError):
        xfer_plan = tx.TransferPlan(
            100,
            lw1.rows()[0][1],
            lw2.rows()[1][1],
            instr_multi,
            max_volume=instr_multi.hw_pipette["working_volume"],
            api_version=ctx.api_version,
            # todo(mm, 2021-03-17): This test intends to test mode='transfer',
            # but it's always accidentally tested mode='consolidate' because of
            # a quirk in how TransferPlan used to guess the mode when not
            # explicitly specified. If this is changed to mode='transfer' now,
            # it raises ZeroDivisionError instead of IndexError. Bug #7516.
            mode="consolidate",
        )
        xfer_plan_list = []
        for step in xfer_plan:
            xfer_plan_list.append(step)


def test_multichannel_transfer_locs(
    hardware: ThreadManagedHardware, deck_definition_name: str
) -> None:
    ctx = papi.create_protocol_context(
        api_version=APIVersion(2, 2),
        hardware_api=hardware,
        deck_type=deck_definition_name,
    )

    lw1 = ctx.load_labware("biorad_96_wellplate_200ul_pcr", 1)
    lw2 = ctx.load_labware("corning_384_wellplate_112ul_flat", 2)
    tiprack = ctx.load_labware("opentrons_96_tiprack_300ul", 3)
    instr_multi = ctx.load_instrument("p300_multi", Mount.LEFT, tip_racks=[tiprack])

    # targets within row limit
    xfer_plan = tx.TransferPlan(
        100,
        lw1.rows()[0][1],
        lw2.rows()[1][1],
        instr_multi,
        max_volume=instr_multi.hw_pipette["working_volume"],
        api_version=ctx.api_version,
        mode="transfer",
    )
    xfer_plan_list = []
    for step in xfer_plan:
        xfer_plan_list.append(step)
    exp1 = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {
            "method": "aspirate",
            "args": [100, lw1.wells_by_name()["A2"], 1.0],
            "kwargs": {},
        },
        {
            "method": "dispense",
            "args": [100, lw2.wells_by_index()["B2"], 1.0],
            "kwargs": {},
        },
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    assert xfer_plan_list == exp1

    # targets outside of row limit will be skipped
    xfer_plan = tx.TransferPlan(
        100,
        lw1.rows()[0][1],
        [lw2.rows()[1][1], lw2.rows()[2][1]],
        instr_multi,
        max_volume=instr_multi.hw_pipette["working_volume"],
        api_version=ctx.api_version,
        mode="transfer",
    )
    xfer_plan_list = []
    for step in xfer_plan:
        xfer_plan_list.append(step)
    assert xfer_plan_list == exp1

    # no valid source or targets, raise error
    with pytest.raises(RuntimeError):
        assert tx.TransferPlan(
            100,
            lw1.rows()[0][1],
            lw2.rows()[2][1],
            instr_multi,
            max_volume=instr_multi.hw_pipette["working_volume"],
            api_version=ctx.api_version,
            mode="transfer",
        )


def test_zero_volume_results_in_no_transfer(_instr_labware: InstrLabware) -> None:
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]
    API_VERSION = _instr_labware["ctx"].api_version

    exp_no_vol = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]

    # ========== Transfer ===========
    xfer_plan = tx.TransferPlan(
        0,
        lw1.columns()[0],
        lw2.columns()[0],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=API_VERSION,
        mode="transfer",
    )
    for step, expected in zip(xfer_plan, exp_no_vol):
        assert step == expected

    xfer_plan = tx.TransferPlan(
        [100, 0, 200],
        lw1.wells()[0:3],
        lw2.wells()[0:3],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=API_VERSION,
        mode="transfer",
    )
    exp2 = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1["A1"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2["A1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [200, lw1["C1"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [200, lw2["C1"], 1.0], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    for step2, expected2 in zip(xfer_plan, exp2):
        assert step2 == expected2

    # ========== Distribute ===========
    dist_plan = tx.TransferPlan(
        0,
        lw1.columns()[0][0],
        lw2.rows()[0][1:3],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=API_VERSION,
        mode="distribute",
    )
    for step, expected in zip(dist_plan, exp_no_vol):
        assert step == expected

    dist_plan = tx.TransferPlan(
        [100, 0],
        lw1.columns()[0][0],
        lw2.rows()[0][1:3],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=API_VERSION,
        mode="distribute",
    )
    exp3 = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1["A1"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2["A2"], 1.0], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    for step2, expected2 in zip(dist_plan, exp3):
        assert step2 == expected2

    # ========== Consolidate ===========
    consd_plan = tx.TransferPlan(
        0,
        lw1.columns()[0],
        lw2.columns()[0][0],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=API_VERSION,
        mode="consolidate",
    )
    for step, expected in zip(consd_plan, exp_no_vol):
        assert step == expected

    cons_list = [100, 200, 300, 200, 0, 0, 100, 200]
    consd_plan = tx.TransferPlan(
        cons_list,
        lw1.columns()[0],
        lw2.columns()[0][0],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=API_VERSION,
        mode="consolidate",
    )
    exp4 = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1["A1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [200, lw1["B1"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [300, lw2["A1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [300, lw1["C1"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [300, lw2["A1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [200, lw1["D1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1["G1"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [300, lw2["A1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [200, lw1["H1"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [200, lw2["A1"], 1.0], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    for step2, expected2 in zip(consd_plan, exp4):
        assert step2 == expected2


def test_zero_volume_causes_transfer_of_disposal_vol(
    _instr_labware: InstrLabware,
) -> None:
    # This test checks the old behavior of distribute and consolidate
    # with zero volumes in which case the volume aspirated/dispensed
    # was the min volume + disposal volume if a zero volume was
    # specified.
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]
    API_VERSION = APIVersion(2, 6)
    assert isinstance(_instr_labware["instr"].trash_container, Labware)
    blow_out = _instr_labware["instr"].trash_container.wells()[0]

    options = tx.TransferOptions()
    options = options._replace(
        transfer=options.transfer._replace(
            disposal_volume=_instr_labware["instr"].min_volume
        )
    )

    # ========== Distribute ===========
    dist_plan = tx.TransferPlan(
        0,
        lw1.columns()[0][0],
        lw2.rows()[0][1:3],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=API_VERSION,
        mode="distribute",
        options=options,
    )

    exp_no_vol = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [30, lw1["A1"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [0, lw2["A2"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [0, lw2["A3"], 1.0], "kwargs": {}},
        {"method": "blow_out", "args": [blow_out], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    for step, expected in zip(dist_plan, exp_no_vol):
        assert step == expected

    dist_plan = tx.TransferPlan(
        [100, 0],
        lw1.columns()[0][0],
        lw2.rows()[0][1:3],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=API_VERSION,
        mode="distribute",
        options=options,
    )
    exp = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [130, lw1["A1"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [100, lw2["A2"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [0, lw2["A3"], 1.0], "kwargs": {}},
        {"method": "blow_out", "args": [blow_out], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    for step, expected in zip(dist_plan, exp):
        assert step == expected

    # ========== Consolidate ===========
    consd_plan = tx.TransferPlan(
        0,
        lw1.columns()[0],
        lw2.columns()[0][0],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=API_VERSION,
        mode="consolidate",
    )

    exp_no_vol = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [0, lw1["A1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [0, lw1["B1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [0, lw1["C1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [0, lw1["D1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [0, lw1["E1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [0, lw1["F1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [0, lw1["G1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [0, lw1["H1"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [0, lw2["A1"], 1.0], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    for step, expected in zip(consd_plan, exp_no_vol):
        assert step == expected

    cons_list = [100, 200, 300, 200, 0, 0, 100, 200]
    consd_plan = tx.TransferPlan(
        cons_list,
        lw1.columns()[0],
        lw2.columns()[0][0],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=API_VERSION,
        mode="consolidate",
    )
    exp2 = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1["A1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [200, lw1["B1"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [300, lw2["A1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [300, lw1["C1"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [300, lw2["A1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [200, lw1["D1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [0, lw1["E1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [0, lw1["F1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [100, lw1["G1"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [300, lw2["A1"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [200, lw1["H1"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [200, lw2["A1"], 1.0], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    for step, expected in zip(consd_plan, exp2):
        assert step == expected


def test_blowout_to_source(_instr_labware: InstrLabware) -> None:
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]
    API_VERSION = APIVersion(2, 6)

    # ========== Transfer ===========
    options = tx.TransferOptions()
    options = options._replace(
        transfer=options.transfer._replace(
            disposal_volume=_instr_labware["instr"].min_volume,
            blow_out_strategy=tx.BlowOutStrategy.SOURCE,
        )
    )

    xfer_plan = tx.TransferPlan(
        30,
        [lw1["A1"], lw1["A2"]],
        [lw2["B1"], lw2["B2"]],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=API_VERSION,
        mode="transfer",
        options=options,
    )

    exp = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [30, lw1["A1"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [30, lw2["B1"], 1.0], "kwargs": {}},
        {"method": "blow_out", "args": [lw1["A1"]], "kwargs": {}},
        {"method": "aspirate", "args": [30, lw1["A2"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [30, lw2["B2"], 1.0], "kwargs": {}},
        {"method": "blow_out", "args": [lw1["A2"]], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    for step, expected in zip(xfer_plan, exp):
        assert step == expected

    # ========== Distribute ===========
    options = tx.TransferOptions()
    options = options._replace(
        transfer=options.transfer._replace(
            disposal_volume=_instr_labware["instr"].min_volume,
            blow_out_strategy=tx.BlowOutStrategy.SOURCE,
        )
    )

    dist_plan = tx.TransferPlan(
        30,
        lw1.columns()[0][0],
        lw2.rows()[0][1:3],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=API_VERSION,
        mode="distribute",
        options=options,
    )

    exp = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [90, lw1["A1"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [30, lw2["A2"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [30, lw2["A3"], 1.0], "kwargs": {}},
        {"method": "blow_out", "args": [lw1["A1"]], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    for step, expected in zip(dist_plan, exp):
        assert step == expected


def test_blowout_to_dest(_instr_labware: InstrLabware) -> None:
    _instr_labware["ctx"].home()
    lw1 = _instr_labware["lw1"]
    lw2 = _instr_labware["lw2"]
    API_VERSION = APIVersion(2, 6)

    # ========== Transfer ===========
    options = tx.TransferOptions()
    options = options._replace(
        transfer=options.transfer._replace(
            disposal_volume=_instr_labware["instr"].min_volume,
            blow_out_strategy=tx.BlowOutStrategy.DEST,
        )
    )

    xfer_plan = tx.TransferPlan(
        30,
        [lw1["A1"], lw1["A2"]],
        [lw2["B1"], lw2["B2"]],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=API_VERSION,
        mode="transfer",
        options=options,
    )

    exp = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [30, lw1["A1"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [30, lw2["B1"], 1.0], "kwargs": {}},
        {"method": "blow_out", "args": [lw2["B1"]], "kwargs": {}},
        {"method": "aspirate", "args": [30, lw1["A2"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [30, lw2["B2"], 1.0], "kwargs": {}},
        {"method": "blow_out", "args": [lw2["B2"]], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    for step, expected in zip(xfer_plan, exp):
        assert step == expected

    # ========== Consolidate ===========
    options = tx.TransferOptions()
    options = options._replace(
        transfer=options.transfer._replace(blow_out_strategy=tx.BlowOutStrategy.DEST)
    )

    consd_plan = tx.TransferPlan(
        30,
        lw2.rows()[0][1:3],
        lw1.columns()[0][0],
        _instr_labware["instr"],
        max_volume=_instr_labware["instr"].hw_pipette["working_volume"],
        api_version=API_VERSION,
        mode="consolidate",
        options=options,
    )

    exp = [
        {"method": "pick_up_tip", "args": [], "kwargs": {}},
        {"method": "aspirate", "args": [30, lw2["A2"], 1.0], "kwargs": {}},
        {"method": "aspirate", "args": [30, lw2["A3"], 1.0], "kwargs": {}},
        {"method": "dispense", "args": [60, lw1["A1"], 1.0], "kwargs": {}},
        {"method": "blow_out", "args": [lw1["A1"]], "kwargs": {}},
        {"method": "drop_tip", "args": [], "kwargs": {}},
    ]
    for step, expected in zip(consd_plan, exp):
        assert step == expected



"""CommandHistory state store tests."""

import pytest

from .command_fixtures import (
    create_queued_command,
)
from opentrons.ordered_set import OrderedSet
from opentrons.protocol_engine.commands import CommandIntent, CommandStatus
from opentrons.protocol_engine.errors.exceptions import CommandDoesNotExistError
from opentrons.protocol_engine.state.command_history import CommandEntry, CommandHistory


def create_queued_command_entry(
    command_id: str = "command-id",
    intent: CommandIntent = CommandIntent.PROTOCOL,
    index: int = 0,
) -> CommandEntry:
    """Create a command entry for a queued command."""
    return CommandEntry(
        create_queued_command(command_id=command_id, intent=intent), index
    )


def create_fixit_command_entry(
    command_id: str = "command-id", index: int = 0
) -> CommandEntry:
    """Create a command entry for a fixit command."""
    return CommandEntry(
        create_queued_command(command_id=command_id, intent=CommandIntent.FIXIT), index
    )


@pytest.fixture
def command_history() -> CommandHistory:
    """Instantiates a CommandHistory instance."""
    return CommandHistory()


def test_length(command_history: CommandHistory) -> None:
    """It should return the length of the command history."""
    assert command_history.length() == 0
    command_history._add("0", create_queued_command_entry())
    assert command_history.length() == 1


def test_has(command_history: CommandHistory) -> None:
    """It should return True if the command exists in the history, False otherwise."""
    assert not command_history.has("0")
    command_history._add("0", create_queued_command_entry())
    assert command_history.has("0")


def test_get(command_history: CommandHistory) -> None:
    """It should return the command entry for the given ID."""
    with pytest.raises(CommandDoesNotExistError):
        command_history.get("0")
    command_entry = create_queued_command_entry()
    command_history._add("0", command_entry)
    assert command_history.get("0") == command_entry


def test_get_next(command_history: CommandHistory) -> None:
    """It should return the next command entry after the command associated with the given ID."""
    with pytest.raises(CommandDoesNotExistError):
        command_history.get_next("0")
    command_entry_1 = create_queued_command_entry()
    command_entry_2 = create_queued_command_entry(index=1)
    command_history._add("0", command_entry_1)
    command_history._add("1", command_entry_2)
    assert command_history.get_next("0") == command_entry_2
    assert command_history.get_next("1") is None


def test_get_prev(command_history: CommandHistory) -> None:
    """It should return the previous command entry before the command associated with the given ID."""
    with pytest.raises(CommandDoesNotExistError):
        command_history.get_prev("0")
    command_entry_1 = create_queued_command_entry()
    command_entry_2 = create_queued_command_entry(index=1)
    command_history._add("0", command_entry_1)
    command_history._add("1", command_entry_2)
    assert command_history.get_prev("0") is None
    assert command_history.get_prev("1") == command_entry_1


def test_get_all_commands(command_history: CommandHistory) -> None:
    """It should return a list of all commands."""
    assert command_history.get_all_commands() == []
    command_entry_1 = create_queued_command_entry()
    command_entry_2 = create_queued_command_entry(index=1)
    command_history._add("0", command_entry_1)
    command_history._add("1", command_entry_2)
    assert command_history.get_all_commands() == [
        command_entry_1.command,
        command_entry_2.command,
    ]


def test_get_all_ids(command_history: CommandHistory) -> None:
    """It should return a list of all command IDs."""
    assert command_history.get_all_ids() == []
    command_entry_1 = create_queued_command_entry()
    command_entry_2 = create_queued_command_entry(index=1)
    command_history._add("0", command_entry_1)
    command_history._add("1", command_entry_2)
    assert command_history.get_all_ids() == ["0", "1"]


def test_get_slice(command_history: CommandHistory) -> None:
    """It should return a slice of commands."""
    assert command_history.get_slice(0, 2) == []
    command_entry_1 = create_queued_command_entry()
    command_entry_2 = create_queued_command_entry(index=1)
    command_entry_3 = create_queued_command_entry(index=2)
    command_history._add("0", command_entry_1)
    command_history._add("1", command_entry_2)
    command_history._add("2", command_entry_3)
    assert command_history.get_slice(1, 3) == [
        command_entry_2.command,
        command_entry_3.command,
    ]


def test_get_tail_command(command_history: CommandHistory) -> None:
    """It should return the tail command."""
    assert command_history.get_tail_command() is None
    command_entry_1 = create_queued_command_entry()
    command_entry_2 = create_queued_command_entry(index=1)
    command_history._add("0", command_entry_1)
    command_history._add("1", command_entry_2)
    assert command_history.get_tail_command() == command_entry_2


def test_get_recently_dequeued_command(command_history: CommandHistory) -> None:
    """It should return the most recently dequeued command."""
    assert command_history.get_most_recently_completed_command() is None
    command_entry = create_queued_command_entry()
    command_history._add("0", command_entry)
    command_history._set_most_recently_completed_command_id("0")
    assert command_history.get_most_recently_completed_command() == command_entry


def test_get_running_command(command_history: CommandHistory) -> None:
    """It should return the currently running command."""
    assert command_history.get_running_command() is None
    command_entry = create_queued_command_entry()
    command_history._add("0", command_entry)
    command_history._set_running_command_id("0")
    assert command_history.get_running_command() == command_entry


def test_get_queue_ids(command_history: CommandHistory) -> None:
    """It should return the IDs of all commands in the queue."""
    assert command_history.get_queue_ids() == OrderedSet()
    command_history._add_to_queue("0")
    command_history._add_to_queue("1")
    assert command_history.get_queue_ids() == OrderedSet(["0", "1"])


def test_get_setup_queue_ids(command_history: CommandHistory) -> None:
    """It should return the IDs of all commands in the setup queue."""
    assert command_history.get_setup_queue_ids() == OrderedSet()
    command_history._add_to_setup_queue("0")
    command_history._add_to_setup_queue("1")
    assert command_history.get_setup_queue_ids() == OrderedSet(["0", "1"])


def test_get_fixit_queue_ids(command_history: CommandHistory) -> None:
    """It should return the IDs of all commands in the setup queue."""
    assert command_history.get_fixit_queue_ids() == OrderedSet()
    command_history._add_to_fixit_queue("0")
    command_history._add_to_fixit_queue("1")
    assert command_history.get_fixit_queue_ids() == OrderedSet(["0", "1"])


def test_set_command_entry(command_history: CommandHistory) -> None:
    """It should set the command entry for the given ID."""
    command_entry = create_queued_command_entry()
    command_history._add("0", command_entry)
    assert command_history.get("0") == command_entry


def test_set_recent_dequeued_command_id(command_history: CommandHistory) -> None:
    """It should set the ID of the most recently dequeued command."""
    command_entry = create_queued_command_entry()
    command_history._add("0", command_entry)
    command_history._set_most_recently_completed_command_id("0")
    assert command_history.get_most_recently_completed_command() == command_entry


def test_set_running_command_id(command_history: CommandHistory) -> None:
    """It should set the ID of the currently running command."""
    command_entry = create_queued_command_entry()
    command_history._add("0", command_entry)
    command_history._set_running_command_id("0")
    assert command_history.get_running_command() == command_entry


def test_set_fixit_running_command_id(command_history: CommandHistory) -> None:
    """It should set the ID of the currently running fixit command."""
    command_entry = create_queued_command()
    command_history.append_queued_command(command_entry)
    running_command = command_entry.model_copy(
        update={
            "status": CommandStatus.RUNNING,
        }
    )
    command_history.set_command_running(running_command)
    finished_command = command_entry.model_copy(
        update={
            "status": CommandStatus.SUCCEEDED,
        }
    )
    command_history.set_command_succeeded(finished_command)
    fixit_command_entry = create_queued_command(
        command_id="fixit-id", intent=CommandIntent.FIXIT
    )
    command_history.append_queued_command(fixit_command_entry)
    fixit_running_command = fixit_command_entry.model_copy(
        update={
            "status": CommandStatus.RUNNING,
        }
    )
    command_history.set_command_running(fixit_running_command)
    current_running_command = command_history.get_running_command()
    assert current_running_command is not None
    assert current_running_command.command == fixit_running_command
    assert command_history.get_all_commands() == [
        finished_command,
        fixit_running_command,
    ]


def test_add_to_queue(command_history: CommandHistory) -> None:
    """It should add the given ID to the queue."""
    command_history._add_to_queue("0")
    assert command_history.get_queue_ids() == OrderedSet(["0"])


def test_add_to_setup_queue(command_history: CommandHistory) -> None:
    """It should add the given ID to the setup queue."""
    command_history._add_to_setup_queue("0")
    assert command_history.get_setup_queue_ids() == OrderedSet(["0"])


def test_add_to_fixit_queue(command_history: CommandHistory) -> None:
    """It should add the given ID to the setup queue."""
    fixit_command = create_queued_command(intent=CommandIntent.FIXIT)
    command_history.append_queued_command(fixit_command)
    assert command_history.get_fixit_queue_ids() == OrderedSet(["command-id"])


def test_remove_id_from_queue(command_history: CommandHistory) -> None:
    """It should remove the given ID from the queue."""
    command_history._add_to_queue("0")
    command_history._add_to_queue("1")
    command_history._remove_queue_id("0")
    assert command_history.get_queue_ids() == OrderedSet(["1"])


def test_remove_id_from_setup_queue(command_history: CommandHistory) -> None:
    """It should remove the given ID from the setup queue."""
    command_history._add_to_setup_queue("0")
    command_history._add_to_setup_queue("1")
    command_history._remove_setup_queue_id("0")
    assert command_history.get_setup_queue_ids() == OrderedSet(["1"])


def test_get_filtered_commands(command_history: CommandHistory) -> None:
    """It should return a list of all commands without fixit commands."""
    assert (
        list(command_history.get_filtered_command_ids(include_fixit_commands=False))
        == []
    )
    command_entry_1 = create_queued_command(command_id="0")
    command_entry_2 = create_queued_command(command_id="1")
    fixit_command_entry_1 = create_queued_command(
        intent=CommandIntent.FIXIT, command_id="fixit-1"
    )
    command_history.append_queued_command(command_entry_1)
    command_history.append_queued_command(command_entry_2)
    command_history.append_queued_command(fixit_command_entry_1)
    assert list(
        command_history.get_filtered_command_ids(include_fixit_commands=False)
    ) == ["0", "1"]


def test_get_all_filtered_commands(command_history: CommandHistory) -> None:
    """It should return a list of all commands without fixit commands."""
    assert (
        list(command_history.get_filtered_command_ids(include_fixit_commands=False))
        == []
    )
    command_entry_1 = create_queued_command_entry()
    command_entry_2 = create_queued_command_entry(index=1, intent=CommandIntent.SETUP)
    fixit_command_entry_1 = create_queued_command_entry(intent=CommandIntent.FIXIT)
    command_history._add("0", command_entry_1)
    command_history._add("1", command_entry_2)
    command_history._add("fixit-1", fixit_command_entry_1)
    assert list(
        command_history.get_filtered_command_ids(include_fixit_commands=True)
    ) == ["0", "1", "fixit-1"]


def test_get_slice_with_filtered_list(command_history: CommandHistory) -> None:
    """It should return a slice of filtered commands."""
    assert command_history.get_slice(0, 2) == []
    command_entry_1 = create_queued_command_entry()
    command_entry_2 = create_queued_command_entry(index=1)
    command_entry_3 = create_queued_command_entry(index=2)
    command_history._add("0", command_entry_1)
    command_history._add("1", command_entry_2)
    command_history._add("2", command_entry_3)
    filtered_list = ["0", "1"]
    assert command_history.get_slice(1, 3, command_ids=filtered_list) == [
        command_entry_2.command,
    ]



"""Common APIs for protocol core interfaces."""

# TODO(mc, 2022-08-22): move to __init__ when dependency cycles are resolved
from .instrument import AbstractInstrument
from .labware import AbstractLabware
from .module import (
    AbstractAbsorbanceReaderCore,
    AbstractFlexStackerCore,
    AbstractHeaterShakerCore,
    AbstractMagneticBlockCore,
    AbstractMagneticModuleCore,
    AbstractModuleCore,
    AbstractTemperatureModuleCore,
    AbstractThermocyclerCore,
    AbstractVacuumModuleCore,
)
from .protocol import AbstractProtocol
from .robot import AbstractRobot
from .tasks import AbstractTaskCore
from .well import AbstractWellCore

WellCore = AbstractWellCore
LabwareCore = AbstractLabware[WellCore]
InstrumentCore = AbstractInstrument[WellCore, LabwareCore]
ModuleCore = AbstractModuleCore[LabwareCore]
TemperatureModuleCore = AbstractTemperatureModuleCore[LabwareCore]
MagneticModuleCore = AbstractMagneticModuleCore[LabwareCore]
ThermocyclerCore = AbstractThermocyclerCore[LabwareCore]
HeaterShakerCore = AbstractHeaterShakerCore[LabwareCore]
MagneticBlockCore = AbstractMagneticBlockCore[LabwareCore]
AbsorbanceReaderCore = AbstractAbsorbanceReaderCore[LabwareCore]
FlexStackerCore = AbstractFlexStackerCore[LabwareCore]
VacuumModuleCore = AbstractVacuumModuleCore[LabwareCore]
RobotCore = AbstractRobot
TaskCore = AbstractTaskCore
ProtocolCore = AbstractProtocol[InstrumentCore, LabwareCore, ModuleCore, TaskCore]
